import numpy as np
from itertools import combinations
from matplotlib import pyplot as plt
from matplotlib.colors import Normalize
import matplotlib.ticker as mticker
from ase.data import covalent_radii
from MCRT.assets.colors import cpk_colors
from MCRT.visualize.utils import plot_cube
from matplotlib.colors import LinearSegmentedColormap
import matplotlib.gridspec as gridspec
from mpl_toolkits.axes_grid1 import make_axes_locatable


plt.rcParams["font.family"] = "Times New Roman"

def draw_colorbar(fig, ax, cmap, minatt, maxatt, **cbar_kwargs):
    norm = Normalize(0.0, 1.0)
    smap = plt.cm.ScalarMappable(cmap=cmap, norm=norm)
    cbar = fig.colorbar(
        smap, ax=ax, fraction=cbar_kwargs["fraction"], shrink=cbar_kwargs["shrink"]
    )
    cbar.ax.tick_params(labelsize=cbar_kwargs["fontsize"])
    ticks_loc = np.linspace(0, 1, cbar_kwargs["num_ticks"])
    ticks_label = np.round(
        np.linspace(minatt, maxatt, cbar_kwargs["num_ticks"]),
        decimals=cbar_kwargs["decimals"],
    )
    cbar.ax.yaxis.set_major_locator(mticker.FixedLocator(ticks_loc))
    cbar.ax.set_yticklabels(ticks_label)

    cbar.ax.set_ylabel(
        "Attention score",
        rotation=270,
        labelpad=cbar_kwargs["labelpad"],
        fontdict={"size": cbar_kwargs["labelsize"]},
    )


def draw_line(ax, pos1, pos2, **kwargs):
    """
    Draw line from position 1 to position 2
    :param ax: <matplotlib.axes> figure axis
    :param pos1: <np.array> starting point position
    :param pos2: <np.array> end point position
    :param kwargs: matplotlib plot3D kwargs
    :return:
    """
    ax.plot3D(*zip(pos1, pos2), **kwargs)


def draw_cell(ax, lattice, s_point=None, **kwargs):
    """
    Draw unit-p_lattice p_lattice using matplotlib
    :param ax: <matplotlib.axes> figure axis
    :param lattice: <np.array> p_lattice vectors (3 X 3 matrix)
    :param s_point: <np.array> start point of p_lattice
    :param kwargs: matplotlib plot3D kwargs
    """
    vec1, vec2, vec3 = lattice
    if s_point is None:
        s_point = np.zeros(3)

    opp_vec = vec1 + vec2 + vec3 + s_point

    for v1, v2 in combinations([vec1, vec2, vec3], 2):
        draw_line(ax, s_point, s_point + v1, **kwargs)
        draw_line(ax, s_point, s_point + v2, **kwargs)
        draw_line(ax, s_point + v1, s_point + v1 + v2, **kwargs)
        draw_line(ax, s_point + v2, s_point + v1 + v2, **kwargs)
        draw_line(ax, s_point + v1 + v2, opp_vec, **kwargs)


def draw_atoms(ax, atoms, atomic_scale):
    """
    Draw p_atoms using matplotlib
    :param ax: <matplotlib.axes> figure axis
    :param atoms: <ase.p_atoms> Target p_atoms for drawing
    :param atomic_scale: <float> scaling factor for draw_atoms.
    """
    coords = atoms.get_positions()
    atomic_numbers = atoms.get_atomic_numbers()
    atomic_sizes = np.array([covalent_radii[i] for i in atomic_numbers])
    atomic_colors = np.array([cpk_colors[i] for i in atomic_numbers])
    ax.scatter(
        xs=coords[:, 0],
        ys=coords[:, 1],
        zs=coords[:, 2],
        c=atomic_colors,
        s=atomic_sizes * atomic_scale,
        marker="o",
        edgecolor="black",
        linewidths=0.8,
        alpha=1.0,
    )


def draw_heatmap_graph(ax, atoms, colors, atomic_scale, alpha):
    coords = atoms.get_positions()
    for i, coord in enumerate(coords):
        ax.scatter(
            xs=coord[0],
            ys=coord[1],
            zs=coord[2],
            color=colors[i],
            s=atomic_scale,
            marker="o",
            linewidth=0,
            alpha=alpha,
        )



def tick_labels(max_birth, max_persistence, pixel_size):
    """Convert image units to units of the persistence diagram.

    Args:
        max_birth: Maximum birth time for the x-axis.
        max_persistence: Maximum persistence time for the y-axis.
        pixel_size: Pixel size resolution for the image (int).
    """

    ticks = np.linspace(0, pixel_size, 6)

    ticklabels_x = [(max_birth / pixel_size) * i for i in ticks]
    ticklabels_y = [(max_persistence / pixel_size) * i for i in ticks]

    ticklabels_x = [round(elem, 2) for elem in ticklabels_x]
    ticklabels_y = [round(elem, 2) for elem in ticklabels_y]

    return ticklabels_x, ticklabels_y


def filter_top_n_heatmap(heatmap, top_n):
    """Keep only the top N elements in the heatmap and set others to zero."""
    flattened_heatmap = heatmap.flatten()
    top_indices = np.argpartition(flattened_heatmap, -top_n)[-top_n:]
    filtered_heatmap = np.zeros_like(flattened_heatmap)
    filtered_heatmap[top_indices] = flattened_heatmap[top_indices]
    return filtered_heatmap.reshape(heatmap.shape)

def get_highest_attention_patch_center(heatmap, max_birth, max_persistence):
    """Get the center coordinates of the highest attention patch."""
    max_index = np.unravel_index(np.argmax(heatmap, axis=None), heatmap.shape)
    patch_size = 5  # Each patch is 5x5 pixels
    center_x = (max_birth / 10) * (max_index[1] + 0.5)
    center_y = (max_persistence / 10) * (10-(max_index[0] + 0.5))
    return center_x, center_y

def create_blue_colormap():
    """Create a blue colormap with transparency for zero values."""
    colors = [(0, 0, 0, 0), (0, 0, 1, 1)]  # RGBA for transparent and blue
    return LinearSegmentedColormap.from_list('blue_cmap', colors, N=256)

def create_pink_colormap():
    """Create a pink colormap with transparency for zero values."""
    colors = [(1, 0.9, 0.9, 0.2), (1, 0.75, 0.8, 1)]  # RGBA for transparent and pink
    return LinearSegmentedColormap.from_list('pink_cmap', colors, N=256)

def create_orange_red_colormap():
    """Create an orange-red colormap with transparency for zero values."""
    colors = color_steps = [(1, 0.9, 0.9, 0)] + [(1, 0.8 - i * (1/30), 0.8 - i * (1/30), 1) for i in range(30)]  # RGBA for transparent, orange, and red
    return LinearSegmentedColormap.from_list('orange_red_cmap', colors, N=256)

def plot_image(image, max_birth, max_persistence, heatmap, top_n=10):
    """Take a persistence image and plot it with a heatmap overlay.

    Args:
        image: The persistence image to be plotted.
        max_birth: Maximum birth time for the x-axis.
        max_persistence: Maximum persistence time for the y-axis.
        heatmap: 1D array representing the heatmap image.
        top_n: Number of top elements to keep in the heatmap.
    """

    # Take the logarithm of the image to enhance contrast
    image_log = np.log1p(image)

    fig, ax = plt.subplots(figsize=(8, 6))

    pixel_size = image.shape[0]  
    ticklabels_x, ticklabels_y = tick_labels(max_birth, max_persistence, pixel_size)
    ticks = np.linspace(0, pixel_size, 6)

    # Display the original image
    im = ax.imshow(image_log, cmap=plt.cm.viridis_r, origin='upper', extent=[0, pixel_size, 0, pixel_size])

    # Prepare the heatmap
    heatmap_reshaped = heatmap[:100].reshape((10, 10))  # Adjust the heatmap size to 10x10
    heatmap_reshaped[heatmap_reshaped < 0.001] = 0  # Set values below 0.001 to zero

    # Get the highest attention patch center
    center_x, center_y = get_highest_attention_patch_center(heatmap_reshaped, max_birth, max_persistence)
    print(f"Highest Attention Patch Center - Birth: {center_x}, Persistence: {center_y}")

    # Filter to keep only top N values
    heatmap_filtered = filter_top_n_heatmap(heatmap_reshaped, top_n)

    # Upsample the filtered heatmap to match the image size (50x50)
    # heatmap_resized = np.kron(heatmap_filtered, np.ones((5, 5)))  # Use np.kron to upsample
    heatmap_resized = np.clip(np.kron(heatmap_filtered, np.ones((5, 5))), 0, 0.03)
    # Create an orange-red colormap
    orange_red_cmap = create_orange_red_colormap()

    # Overlay the heatmap
    # hm = ax.imshow(heatmap_resized, cmap=orange_red_cmap, alpha=0.5, extent=[0, pixel_size, 0, pixel_size])
    hm = ax.imshow(heatmap_resized, cmap=orange_red_cmap, alpha=0.4, extent=[0, pixel_size, 0, pixel_size], vmin=0, vmax=0.03)
    # Set tick labels
    ax.set_xticks(ticks)
    ax.set_yticks(ticks)
    ax.set_xticklabels(ticklabels_x)
    ax.set_yticklabels(ticklabels_y)
    ax.set_xlabel("Birth",fontsize=16)
    ax.set_ylabel("Persistence",fontsize=16)
    # ax.set_title("Persistence Image")

    # Add colorbars
    divider = make_axes_locatable(ax)
    cax_im = divider.append_axes("right", size="5%", pad=0.1)
    cax_hm = divider.append_axes("right", size="5%", pad=0.46)  # Adjust pad to move cax_hm to the left

    cbar_image = plt.colorbar(im, cax=cax_im)
    cbar_image.set_label('Image Intensity')

    cbar_heatmap = plt.colorbar(hm, cax=cax_hm)
    cbar_heatmap.set_label('Attention Score')

    plt.tight_layout()
    return fig,ax
    # plt.show()



import torch
from torchmetrics import Metric


class Accuracy(Metric):
    def __init__(self, dist_sync_on_step=False):
        super().__init__(dist_sync_on_step=dist_sync_on_step)
        self.add_state("correct", default=torch.tensor(0.0), dist_reduce_fx="sum")
        self.add_state("total", default=torch.tensor(0.0), dist_reduce_fx="sum")

    def update(self, logits, target):
        logits, target = (
            logits.detach().to(self.correct.device),
            target.detach().to(self.correct.device),
        )

        if len(target.shape) > 1 and target.shape[1] > 1:  # muilti-hot
            preds = torch.sigmoid(logits) > 0.5  
            correct = torch.sum((preds == target).all(dim=1)).float() # only all correct can be considered correct prediction
            total = target.size(0)
        else:
            if len(logits.shape) > 1:
                preds = logits.argmax(dim=-1)
            else:  # binary accuracy
                preds = (logits >= 0.5).float()

            preds = preds[target != -1]  # invalid samples = -1
            target = target[target != -1]

            if target.numel() == 0:
                return 1  

            correct = torch.sum(preds == target).float()
            total = target.numel()
            
        self.correct += correct
        self.total += total

    def compute(self):
        return self.correct / self.total if self.total > 0 else torch.tensor(0.0)


class Scalar(Metric):
    def __init__(self, dist_sync_on_step=False):
        super().__init__(dist_sync_on_step=dist_sync_on_step)
        self.add_state("scalar", default=torch.tensor(0.0), dist_reduce_fx="sum")
        self.add_state("total", default=torch.tensor(0.0), dist_reduce_fx="sum")

    def update(self, scalar):
        if isinstance(scalar, torch.Tensor):
            scalar = scalar.detach().to(self.scalar.device)
        else:
            scalar = torch.tensor(scalar).float().to(self.scalar.device)
        self.scalar += scalar
        self.total += 1

    def compute(self):
        return self.scalar / self.total


import os
from MCRT import __root_dir__

DEFAULT_PRETRAIN_MODEL_PATH = os.path.join(
    __root_dir__, "models/"
)

DEFAULT_MCRT_PATH = os.path.join(__root_dir__, "models/MCRT.ckpt")

DEFAULT_FINETUNED_MODEL_PATH = os.path.join(__root_dir__, "models/finetuned/")



from typing import Tuple, Any, Dict

import dgl
import dgl.function as fn
import torch
import torch.nn as nn
import torch.nn.functional as F
from typing import Optional
import numpy as np

class MLPReadout(nn.Module):
    """
    Readout function.
    """

    def __init__(
        self,
        input_dim: int,
        output_dim: int,
        L: int = 2,
    ):  # L=nb_hidden_layers
        super().__init__()
        list_FC_layers = [
            nn.Linear(input_dim // 2**l, input_dim // 2 ** (l + 1), bias=True)
            for l in range(L)
        ]
        list_FC_layers.append(nn.Linear(input_dim // 2**L, output_dim, bias=True))
        self.FC_layers = nn.ModuleList(list_FC_layers)
        self.L = L

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        y = x
        for l in range(self.L):
            y = self.FC_layers[l](y)
            y = F.relu(y)
        y = self.FC_layers[self.L](y)
        return y
class GatedGCNLayer(nn.Module):
    """ResGatedGCN: Residual Gated Graph ConvNets

    "An Experimental Study of Neural Networks for Variable Graph"
    ICLR (2018)
    https://arxiv.org/pdf/1711.07553v2.pdf
    """

    def __init__(
        self,
        input_dim,
        output_dim,
        batch_norm: bool = False,
        residual: bool = False,
        dropout: float = 0.0,
    ):
        super().__init__()
        self.input_dim = input_dim
        self.output_dim = output_dim
        self.dropout = dropout
        self.batch_norm = batch_norm
        self.residual = residual

        if input_dim != output_dim:
            self.residual = False

        self.A = nn.Linear(input_dim, output_dim)
        self.B = nn.Linear(input_dim, output_dim)
        self.C = nn.Linear(input_dim, output_dim)
        self.D = nn.Linear(input_dim, output_dim)
        self.E = nn.Linear(input_dim, output_dim)
        self.bn_node_h = nn.BatchNorm1d(output_dim)
        self.bn_node_e = nn.BatchNorm1d(output_dim)

    def forward(
        self,
        g: dgl.DGLGraph,
        h: torch.Tensor,
        e: torch.Tensor,
    ):
        """
        Args:
            g (dgl.DGLGraph): DGLGraph
            h (torch.Tensor): embedded node features
            e (torch.Tensor): embedded edge features
        Return:
            h (torch.Tensor): updated node features
            e (torch.Tensor): updated edge features
        """
        h_in = h  # for residual connection
        e_in = e  # for residual connection

        g.ndata["h"] = h
        g.ndata["Ah"] = self.A(h)
        g.ndata["Bh"] = self.B(h)
        g.ndata["Dh"] = self.D(h)
        g.ndata["Eh"] = self.E(h)
        g.edata["e"] = e
        g.edata["Ce"] = self.C(e)

        g.apply_edges(fn.u_add_v("Dh", "Eh", "DEh"))
        g.edata["e"] = g.edata["DEh"] + g.edata["Ce"]  # updated edge features (e^_ij)
        g.edata["sigma"] = torch.sigmoid(g.edata["e"])  # sigma(e^_ij)
        # numerator
        g.update_all(fn.u_mul_e("Bh", "sigma", "m"), fn.sum("m", "sum_sigma_h"))
        # denominator
        g.update_all(fn.copy_e("sigma", "m"), fn.sum("m", "sum_sigma"))
        g.ndata["h"] = g.ndata["Ah"] + g.ndata["sum_sigma_h"] / (
            g.ndata["sum_sigma"] + 1e-6
        )  # updated node features
        h = g.ndata["h"]
        e = g.edata["e"]

        if self.batch_norm:
            h = self.bn_node_h(h)
            e = self.bn_node_e(e)

        h = F.silu(h)
        e = F.silu(e)

        if self.residual:
            h = h_in + h
            e = e_in + e

        h = F.dropout(h, self.dropout, training=self.training)
        e = F.dropout(e, self.dropout, training=self.training)

        return h, e
class ALIGNNLayer(nn.Module):
    """ALIGNN layer.

    "Atomistic Line Graph Neural Network
    for improved materials property predictions"
    npj Comput. Mater. (2021).
    https://www.nature.com/articles/s41524-021-00650-1
    """

    def __init__(
        self,
        input_dim: int,
        output_dim: int,
        batch_norm: bool = False,
        residual: bool = False,
        dropout: float = 0.0,
    ):
        """
        Args:
            input_dim (int): dimension of input features
            output_dim (int): dimension of output features
            batch_norm (bool, optional): whether to use batch normalization. Defaults to False.
            residual (bool, optional): whether to use residual connection. Defaults to False.
            dropout (float, optional): a ratio of dropout. Defaults to 0.0.
        """
        super().__init__()
        self.node_update = GatedGCNLayer(
            input_dim=input_dim,
            output_dim=output_dim,
            batch_norm=batch_norm,
            residual=residual,
            dropout=dropout,
        )
        self.edge_update = GatedGCNLayer(
            input_dim=output_dim,
            output_dim=input_dim,
            batch_norm=batch_norm,
            residual=residual,
            dropout=dropout,
        )

    def forward(
        self,
        g: dgl.DGLGraph,
        lg: dgl.DGLGraph,
        h: torch.Tensor,
        e: torch.Tensor,
        l: torch.Tensor,
    ) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        """
        Args:
            g (dgl.DGLGraph): DGLGraph
            lg (dgl.DGLGraph): line graph of g
            h (torch.Tensor): embedded node features
            e (torch.Tensor): embedded edge features
            l (torch.Tensor): embedded line graph features (edge pair)

        Returns:
            Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
            updated node features, updated edge features, updated line graph features
        """
        # node update
        h, m = self.node_update(g, h, e)
        # edge update
        e, l = self.edge_update(lg, m, l)

        return h, e, l
class RBFExpansion(nn.Module):
    """Expand interatomic distances with radial basis functions."""

    def __init__(
        self,
        vmin: float = 0,
        vmax: float = 8,
        bins: int = 40,
        lengthscale: Optional[float] = None,
    ):
        """Register torch parameters for RBF expansion."""
        super().__init__()
        self.vmin = vmin
        self.vmax = vmax
        self.bins = bins
        self.register_buffer("centers", torch.linspace(self.vmin, self.vmax, self.bins))

        if lengthscale is None:
            # SchNet-style
            # set lengthscales relative to granularity of RBF expansion
            self.lengthscale = np.diff(self.centers).mean()
            self.gamma = 1 / self.lengthscale

        else:
            self.lengthscale = lengthscale
            self.gamma = 1 / (lengthscale**2)

    @torch.no_grad()
    def forward(self, distance: torch.Tensor) -> torch.Tensor:
        """Apply RBF expansion to interatomic distance tensor."""
        return torch.exp(-self.gamma * (distance.unsqueeze(1) - self.centers) ** 2)


class ALIGNN(nn.Module):
    """ALIGNN model. without pooling and readout

    "Atomistic Line Graph Neural Network
    for improved materials property predictions"
    npj Comput. Mater. (2021).
    https://www.nature.com/articles/s41524-021-00650-1
    """

    def __init__(self, num_conv,hidden_dim,rbf_distance_dim,rbf_triplet_dim,batch_norm,dropout,residual,transformer_hidden_dim,max_graph_len ):
        super().__init__()
        # config
        self.num_conv = num_conv
        self.hidden_dim = hidden_dim
        self.rbf_distance_dim = rbf_distance_dim
        self.rbf_triplet_dim = rbf_triplet_dim
        self.batch_norm = batch_norm
        self.dropout = dropout
        self.residual = residual
        self.transformer_hidden_dim = transformer_hidden_dim
        self.max_graph_len = max_graph_len 

        # layers
        self.node_embedding = nn.Embedding(103, self.hidden_dim)
        self.edge_embedding = nn.Linear(self.rbf_distance_dim, self.hidden_dim)
        self.angle_embedding = nn.Linear(self.rbf_triplet_dim, self.hidden_dim)
        self.rbf_expansion_distance = RBFExpansion(
            vmin=0, vmax=8, bins=self.rbf_distance_dim
        )
        self.rbf_expansion_triplet = RBFExpansion(
            vmin=-1, vmax=1, bins=self.rbf_triplet_dim
        )
        self.conv_layers = nn.ModuleList(
            [
                ALIGNNLayer(
                    input_dim=self.hidden_dim,
                    output_dim=self.hidden_dim,
                    batch_norm=self.batch_norm,
                    residual=self.residual,
                    dropout=self.dropout,
                )
                for _ in range(self.num_conv)
            ]
        )
        self.gated_gcn_layers = nn.ModuleList(
            [
                GatedGCNLayer(
                    input_dim=self.hidden_dim,
                    output_dim=self.hidden_dim,
                    batch_norm=self.batch_norm,
                    residual=self.residual,
                    dropout=self.dropout,
                )
                for _ in range(self.num_conv)
            ]
        )
        self.fc = nn.Linear(self.hidden_dim, self.transformer_hidden_dim)

    def forward(self, batch: Dict[str, Any]) -> torch.Tensor:
        """Forward propagation.

        Args:
            batch (Dict[str, Any]): batch data including graph, line graph (optional)
            and target

        Returns:
            torch.Tensor: predicted target values (logits)
        """
        graph = batch["graph"]
        line_graph = batch["line_graph"]
        crystal_atom_idx = batch["crystal_atom_idx"]

        # node embedding
        node_attrs = graph.ndata["atomic_number"]
        node_feats = self.node_embedding(node_attrs)
        # edge embedding
        edge_attrs = self.rbf_expansion_distance(graph.edata["distance"])
        edge_feats = self.edge_embedding(edge_attrs)
        # angle (edge pair) embedding
        angle_attrs = self.rbf_expansion_triplet(line_graph.edata["angle"])
        angle_feats = self.angle_embedding(angle_attrs)
        # conv layers
        for conv in self.conv_layers:
            node_feats, edge_feats, angle_feats = conv(
                graph, line_graph, node_feats, edge_feats, angle_feats
            )
        # gated gcn layers
        for gated_gcn in self.gated_gcn_layers:
            node_feats, _ = gated_gcn(graph, node_feats, edge_feats)
        # adapt dim    
        node_feats = self.fc(node_feats)  
        new_node_feats, mask = self.reconstruct_batch(node_feats, crystal_atom_idx)  
        return new_node_feats, mask
    
    def reconstruct_batch(self, atom_fea, crystal_atom_idx):
        batch_size = len(crystal_atom_idx)
        new_atom_fea = torch.zeros(batch_size, self.max_graph_len, self.transformer_hidden_dim).to(atom_fea)
        mask_label = torch.zeros(batch_size, self.max_graph_len).to(atom_fea) # this mask is for label valid position
        for i, idx in enumerate(crystal_atom_idx):
                length = len(idx)
                new_atom_fea[i, :length] = atom_fea[idx]
                mask_label[i, :length] = 1  # valid position set to 1

        return new_atom_fea, mask_label # this mask is for label valid position


""" 
MCRT_Transformer
"""
from functools import partial
import torch
import torch.nn as nn
from torch.nn import AvgPool3d
from timm.models.layers import DropPath, trunc_normal_

#test
from MCRT.data_processor.dataset import Dataset
from MCRT.modules.cgcnn import GraphEmbeddings
from einops.layers.torch import Rearrange

def main():
    dataset_test=Dataset(r'D:\Projects\MyProjects\MCRT\MCRT\cifs\test','train',1023,4)
    batch=[dataset_test[i] for i in range(32)]
    collated_batch=dataset_test.collate(batch)
    # print(collated_batch["padded_distance_matrices"].shape)
    GraphEmbeddingsmodel=GraphEmbeddings(atom_fea_len=64, nbr_fea_len=41, max_graph_len=1023, hid_dim=512,is_finetune=False, n_conv=3,mask_probability=0.15)
    new_atom_fea, atom_label,atm_label, mask=GraphEmbeddingsmodel(atom_num=collated_batch["atom_num"], nbr_idx=collated_batch["nbr_idx"], nbr_fea=collated_batch["nbr_fea"], \
                                crystal_atom_idx=collated_batch["crystal_atom_idx"], atm_list=collated_batch["atm_list"])
    McrtTransformer=MCRT_Transformer(
        num_blocks=12,
        dim=512,
        num_heads=8,
        mlp_ratio=4,
        qkv_bias=True,
        qk_scale=None,
        drop=0.0,
        attn_drop=0.0,
        drop_path_rate=0.1,
        act_layer=nn.GELU,
        norm_layer=nn.LayerNorm
    )
    new_atom_fea=McrtTransformer(new_atom_fea,mask)
    print(new_atom_fea)

class Mlp(nn.Module):
    def __init__(
        self,
        in_features,
        hidden_features=None,
        out_features=None,
        act_layer=nn.GELU,
        drop=0.0,
    ):
        super().__init__()
        out_features = out_features or in_features
        hidden_features = hidden_features or in_features
        self.fc1 = nn.Linear(in_features, hidden_features)
        self.act = act_layer()
        self.fc2 = nn.Linear(hidden_features, out_features)
        self.drop = nn.Dropout(drop)

    def forward(self, x):
        x = self.fc1(x)
        x = self.act(x)
        x = self.drop(x)
        x = self.fc2(x)
        x = self.drop(x)
        return x


class Attention(nn.Module):
    def __init__(
        self,
        dim,
        num_heads=8,
        qkv_bias=False,
        qk_scale=None,
        attn_drop=0.0,
        proj_drop=0.0,
    ):
        super().__init__()
        self.num_heads = num_heads
        head_dim = dim // num_heads
        self.scale = qk_scale or head_dim**-0.5

        self.qkv = nn.Linear(dim, dim * 3, bias=qkv_bias)
        self.attn_drop = nn.Dropout(attn_drop)
        self.proj = nn.Linear(dim, dim)
        self.proj_drop = nn.Dropout(proj_drop)

    def forward(self, x, mask=None):
        B, N, C = x.shape
        assert C % self.num_heads == 0
        qkv = (
            self.qkv(x)  # [B, N, 3*C]
            .reshape(
                B, N, 3, self.num_heads, C // self.num_heads
            )  # [B, N, 3, num_heads, C//num_heads]
            .permute(2, 0, 3, 1, 4)  # [3, B, num_heads, N, C//num_heads]
        )
        q, k, v = (
            qkv[0],  # [B, num_heads, N, C//num_heads]
            qkv[1],  # [B, num_heads, N, C//num_heads]
            qkv[2],  # [B, num_heads, N, C//num_heads]
        )

        attn = (q @ k.transpose(-2, -1)) * self.scale  # [B, num_heads, N, N]
        if mask is not None:
            mask = mask.bool()
            attn = attn.masked_fill(~mask[:, None, None, :], float("-inf"))
        attn = attn.softmax(dim=-1)  # [B, num_heads, N, N]
        attn = self.attn_drop(attn)

        x = (
            (attn @ v).transpose(1, 2).reshape(B, N, C)
        )  # [B, num_heads, N, C//num_heads] -> [B, N, C]
        x = self.proj(x)
        x = self.proj_drop(x)
        return x, attn


class Block(nn.Module):
    def __init__(
        self,
        dim,
        num_heads,
        mlp_ratio=4.0,
        qkv_bias=False,
        qk_scale=None,
        drop=0.0,
        attn_drop=0.0,
        drop_path=0.0,
        act_layer=nn.GELU,
        norm_layer=nn.LayerNorm,
    ):
        super().__init__()
        self.norm1 = norm_layer(dim)
        self.attn = Attention(
            dim,
            num_heads=num_heads,
            qkv_bias=qkv_bias,
            qk_scale=qk_scale,
            attn_drop=attn_drop,
            proj_drop=drop,
        )
        self.drop_path = DropPath(drop_path) if drop_path > 0.0 else nn.Identity()
        self.norm2 = norm_layer(dim)
        mlp_hidden_dim = int(dim * mlp_ratio)
        self.mlp = Mlp(
            in_features=dim,
            hidden_features=mlp_hidden_dim,
            act_layer=act_layer,
            drop=drop,
        )

    def forward(self, x, mask=None):
        _x, attn = self.attn(self.norm1(x), mask=mask)
        x = x + self.drop_path(_x)
        x = x + self.drop_path(self.mlp(self.norm2(x)))
        return x, attn

class MCRT_Transformer(nn.Module):
    def __init__(
        self,
        num_blocks,
        dim,
        num_heads,
        img_size, 
        patch_size, 
        in_chans=3, 
        mlp_ratio=4.0,
        qkv_bias=True,
        qk_scale=None,
        drop=0.0,
        attn_drop=0.0,
        drop_path_rate=0.1,
        act_layer=nn.GELU,
        norm_layer=nn.LayerNorm
    ):
        super().__init__()

        self.patch_embed_1d = PatchEmbed2D(
            img_size=img_size,
            patch_size=patch_size,
            in_chans=in_chans,
            embed_dim=dim,
        )
        self.patch_embed_2d = PatchEmbed2D(
            img_size=img_size,
            patch_size=patch_size,
            in_chans=in_chans,
            embed_dim=dim,
        )
        self.blocks = nn.ModuleList([
            Block(
                dim=dim,
                num_heads=num_heads,
                mlp_ratio=mlp_ratio,
                qkv_bias=qkv_bias,
                qk_scale=qk_scale,
                drop=drop,
                attn_drop=attn_drop,
                drop_path=drop_path_rate * i / (num_blocks - 1),
                act_layer=act_layer,
                norm_layer=norm_layer
            )
            for i in range(num_blocks)
        ])
        self.norm = norm_layer(dim)
        self.apply(self._init_weights)

    def forward(self, x, mask=None):
        for block in self.blocks:
            x, _ = block(x, mask=mask)
        x = self.norm(x)
        return x
    
    def _init_weights(self, m):
        if isinstance(m, nn.Linear):
            trunc_normal_(m.weight, std=0.02)
            if isinstance(m, nn.Linear) and m.bias is not None:
                nn.init.constant_(m.bias, 0)
        elif isinstance(m, nn.LayerNorm):
            nn.init.constant_(m.bias, 0)
            nn.init.constant_(m.weight, 1.0)

    def image_embed_1d(self, _x,):
        """

        :param _x: batch images, Tensor [B, C, H, W]
        :return:
            x:  Tensor [B, num_patches, hid_dim],
        """

        B, _, _, _,= _x.shape
        x = self.patch_embed_1d(_x)  # [B, num_patches, embed_dim]
        x_mask = torch.ones(x.shape[:2]).to(x)  # [B, num_patches]
        return x,x_mask
    
    def image_embed_2d(self, _x,):
        """

        :param _x: batch images, Tensor [B, C, H, W]
        :return:
            x:  Tensor [B, num_patches, hid_dim],
        """

        B, _, _, _,= _x.shape
        x = self.patch_embed_2d(_x)  # [B, num_patches, embed_dim]
        x_mask = torch.ones(x.shape[:2]).to(x)  # [B, num_patches]
        return x,x_mask
    
def posemb_sincos_2d(h, w, dim, temperature: int = 10000, dtype = torch.float32):
    y, x = torch.meshgrid(torch.arange(h), torch.arange(w), indexing="ij")
    assert (dim % 4) == 0, "feature dimension must be multiple of 4 for sincos emb"
    omega = torch.arange(dim // 4) / (dim // 4 - 1)
    omega = 1.0 / (temperature ** omega)

    y = y.flatten()[:, None] * omega[None, :]
    x = x.flatten()[:, None] * omega[None, :]
    pe = torch.cat((x.sin(), x.cos(), y.sin(), y.cos()), dim=1)
    return pe.type(dtype)


class PatchEmbed2D(nn.Module):
    """ 2D Image to Patch Embedding """

    def __init__(self, img_size, patch_size, in_chans=1, embed_dim=768):
        """
        Args:
            img_size (int or tuple): size of the image (height, width)
            patch_size (int or tuple): size of each patch (patch_height, patch_width)
            in_chans (int): number of input channels (e.g., 3 for RGB images)
            embed_dim (int): dimension of the embedding vectors
        """
        super().__init__()
        if isinstance(img_size, int):
            self.img_size = (img_size, img_size)  # (height, width)
        else:
            self.img_size = img_size

        if isinstance(patch_size, int):
            self.patch_size = (patch_size, patch_size)  # (patch_height, patch_width)
        else:
            self.patch_size = patch_size

        self.in_chans = in_chans
        self.embed_dim = embed_dim

        self.num_patches = (self.img_size[0] // self.patch_size[0]) * (self.img_size[1] // self.patch_size[1])
        patch_dim =self.patch_size[0] * self.patch_size[1] * self.in_chans
        self.proj = nn.Sequential(
            Rearrange('b c (h p1) (w p2) -> b (h w) (p1 p2 c)', p1=self.patch_size[0], p2=self.patch_size[1]),
            # nn.LayerNorm(patch_dim),
            nn.Linear(patch_dim, self.embed_dim),
            # nn.LayerNorm(self.embed_dim),
        )
        
        self.pos_embedding = posemb_sincos_2d(
            h = self.img_size[0] // self.patch_size[0],
            w = self.img_size[1] // self.patch_size[1],
            dim = embed_dim,
        ) 

    def forward(self, x):
        """
        Args:
            x (tensor): Tensor, shape [batch_size, channels, height, width]

        Returns:
            tensor: Tensor, shape [batch_size, num_patches, embed_dim]
        """
        x = self.proj(x)  # Apply the projection to flatten and embed
        x += self.pos_embedding.to(x.device, dtype=x.dtype)
        return x

if __name__ == "__main__":
    main()  


import os
import numpy as np
import csv
from pathlib import Path
from collections.abc import Iterable, Sequence
from itertools import product
from functools import wraps, partial
from matplotlib import pyplot as plt
from matplotlib.colors import ListedColormap
from matplotlib import animation

from MCRT.visualize.utils import (
    get_structure,
    get_heatmap,
    scaler,
    get_model_and_datamodule,
    get_batch_from_index,
    get_batch_from_cif_id,
)
from MCRT.visualize.setting import (
    get_fig_ax,
    set_fig_ax,
    set_axes_equal,
    DEFAULT_FIGSIZE,
    DEFAULT_VIEW_INIT,
    get_default_cbar_kwargs,
    get_cmap,
)
from MCRT.visualize.drawer import (
    draw_cell,
    draw_atoms,
    draw_colorbar,
    draw_heatmap_graph,
    plot_image
)


class PatchVisualizer(object):
    def __init__(self, path_cif, heatmap_graph, heatmap_image_1d,heatmap_image_2d,image_1d,image_2d,max_birth_1d,max_persistence_1d,max_birth_2d,max_persistence_2d, **kwargs):
        """
        Attention Visualizer from MCRT
        :param path_cif: <str> path for original cif file.
        :param heatmap_graph: <np.array> graph attention score for cif
        :param heatmap_grid: <np.array> grid attention score for cif
        :param kwargs:
            figsize : (float, float) figure size
            view_init : (float, float) view init from matplotlib
            show_axis : <bool> If True, axis are visible. (default : False)
            show_colorbar : <bool> If True, colorbar are visible. (default : True)
            cmap : (str or matplotlib.colors.ListedColormap) color map used in figure. (default : None)
            num_patches : (int, int, int) number of patches (default : (6, 6, 6))
            max_length : <float> max p_lattice length of structure file (Å)
            min_length: <float> min p_lattice length of structure file (Å)
        """
        self.path_cif = path_cif
        self.heatmap_graph = heatmap_graph
        self.heatmap_image_1d = heatmap_image_1d
        self.heatmap_image_2d = heatmap_image_2d

        self.image_1d=image_1d
        self.image_2d=image_2d
        self.max_birth_1d=max_birth_1d
        self.max_persistence_1d=max_persistence_1d
        self.max_birth_2d=max_birth_2d
        self.max_persistence_2d=max_persistence_2d

        self.kwargs = kwargs
        self.cbar_kwargs = get_default_cbar_kwargs(self.figsize)

        # get primitive structure from cif
        p_atoms = get_structure(self.path_cif, make_supercell=False, dtype="ase")
        self.p_atoms = p_atoms
        self.p_lattice = p_atoms.cell.array

        # get supercell atoms from cif
        max_length, min_length = kwargs.get("max_length", 60), kwargs.get(
            "min_length", 30
        )
        super_atoms = get_structure(
            self.path_cif,
            make_supercell=True,
            dtype="ase",
            max_length=max_length,
            min_length=min_length,
        )
        self.s_atoms = super_atoms
        self.s_lattice = super_atoms.cell.array

    @classmethod
    def from_batch(cls, batch, batch_idx, model, cif_root,save_heatmap=False, **kwargs):
        """
        Attention visualizer from "MOFTransformer" model and "dataloader batch"
        :param batch: Dataloader -> batch
        :param batch_idx: index for batch index
        :param model: <torch.model> fine-tuned MOFTransformer model
        :param cif_root: <str> root dir for cif files
        :param kwargs:
            figsize : (float, float) figure size
            view_init : (float, float) view init from matplotlib
            show_axis : <bool> If True, axis are visible. (default : False)
            show_colorbar : <bool> If True, colorbar are visible. (default : True)
            cmap : (str or matplotlib.colors.ListedColormap) color map used in figure. (default : None)
            num_patches : (int, int, int) number of patches (default : (6, 6, 6))
            max_length : <float> max p_lattice length of structure file (Å)
            min_length: <float> min p_lattice length of structure file (Å)
        :return: <PatchVisualizer> patch visualizer object
        """
        output = model.infer(batch)
        heatmap_graph, heatmap_image_1d, heatmap_image_2d ,image_1d,image_2d,max_birth_1d,max_persistence_1d,max_birth_2d,max_persistence_2d,= get_heatmap(output, batch_idx)
        if save_heatmap:
            scaled_heatmap_graph = scaler(heatmap_graph, min_att=0.000, max_att=0.010)
            with open('MCRT/visualize/results/atom_attention_weights.csv', 'w', newline='') as file:
                writer = csv.writer(file)
                for index, weight in enumerate(scaled_heatmap_graph):
                    writer.writerow([index, weight])
            with open('MCRT/visualize/results/heatmap_image_1d_attention.csv', 'w', newline='') as file:
                writer = csv.writer(file)
                for index, weight in enumerate(heatmap_image_1d):
                    writer.writerow([index, weight])
            with open('MCRT/visualize/results/heatmap_image_2d_attention.csv', 'w', newline='') as file:
                writer = csv.writer(file)
                for index, weight in enumerate(heatmap_image_2d):
                    writer.writerow([index, weight])

        cif_id = batch["cif_id"][batch_idx]
        path_cif = os.path.join(cif_root, cif_id + ".cif")
        return cls(
            path_cif, heatmap_graph, heatmap_image_1d, heatmap_image_2d,image_1d,image_2d,max_birth_1d,max_persistence_1d,max_birth_2d,max_persistence_2d, cif_id=cif_id, **kwargs
        )

    @classmethod
    def from_index(
        cls, index, model_path, data_root, downstream="", cif_root=None, **kwargs
    ):
        """
        Create PatchVisualizer from index. The index corresponds 1:1 to the MOF in the json file in dataset folder.
        The index matches the cif order of json.
        :param index: (int) index of dataset.
        :param model_path: (str) path of model from fine-tuned MOFTransformer with format '.ckpt'
        :param data_root: (str) path of dataset directory obtained from 'prepared_data.py. (see Dataset Preparation)
                MOFs to be visualized must exist in {dataset_folder}/test.json or {dataset_folder}/test_{downstream}.json,
                and {dataset_folder}/test folder. *.graphdata, *.grid, *.griddata16 files should be existed in {dataset_folder}/test folder.
        :param downstream: (str, optional) Use if data are existed in {dataset_folder}/test_{downstream}.json (default:'')
        :param cif_root: (str, optional) path of directory including cif file. The cif lists in the dataset folder should be included.
                If not specified, it is automatically specified as a {dataset_folder}/test folder.
        :param kwargs:
            figsize : (float, float) figure size
            view_init : (float, float) view init from matplotlib
            show_axis : <bool> If True, axis are visible. (default : False)
            show_colorbar : <bool> If True, colorbar are visible. (default : True)
            cmap : (str or matplotlib.colors.ListedColormap) color map used in figure. (default : None)
            num_patches : (int, int, int) number of patches (default : (6, 6, 6))
            max_length : <float> max p_lattice length of structure file (Å)
            min_length: <float> min p_lattice length of structure file (Å)
        :return: PatchVisualizer class for index
        """
        model, data_iter = get_model_and_datamodule(model_path, data_root, downstream)
        batch = get_batch_from_index(data_iter, index)
        if cif_root is None:
            cif_root = os.path.join(data_root, "test")

        return cls.from_batch(batch, 0, model, cif_root, **kwargs)

    @classmethod
    def from_cifname(
        cls, cifname, model_path, data_root, downstream="", cif_root=None, save_heatmap=False, **kwargs
    ):
        """
        Create PatchVisualizer from cif name. cif must be in test.json or test_{downstream}.json.

        :param cifname : (str) name or path of cif. Data matching the corresponding cif name is retrieved from the dataset.
        :param model_path: (str) path of model from fine-tuned MOFTransformer with format '.ckpt'
        :param data_root: (str) path of dataset directory obtained from 'prepared_data.py. (see Dataset Preparation)
                MOFs to be visualized must exist in {dataset_folder}/test.json or {dataset_folder}/test_{downstream}.json,
                and {dataset_folder}/test folder. *.graphdata, *.grid, *.griddata16 files should be existed in {dataset_folder}/test folder.
        :param downstream: (str, optional) Use if data are existed in {dataset_folder}/test_{downstream}.json (default:'')
        :param cif_root: (str, optional) path of directory including cif file. The cif lists in the dataset folder should be included.
                If not specified, it is automatically specified as a {dataset_folder}/test folder.
        :param kwargs:
            figsize : (float, float) figure size
            view_init : (float, float) view init from matplotlib
            show_axis : <bool> If True, axis are visible. (default : False)
            show_colorbar : <bool> If True, colorbar are visible. (default : True)
            cmap : (str or matplotlib.colors.ListedColormap) color map used in figure. (default : None)
            num_patches : (int, int, int) number of patches (default : (6, 6, 6))
            max_length : <float> max p_lattice length of structure file (Å)
            min_length: <float> min p_lattice length of structure file (Å)
        :return: PatchVisualizer class for index
        """
        model, data_iter = get_model_and_datamodule(model_path, data_root, downstream)
        batch = get_batch_from_cif_id(data_iter, cifname)
        if cif_root is None:
            cif_root = os.path.join(data_root, "test")
        if save_heatmap:
            return cls.from_batch(batch, 0, model, cif_root,True, **kwargs)
        else:
            return cls.from_batch(batch, 0, model, cif_root, **kwargs)

    def __repr__(self):
        return f"class <PatchVisualizer> from {self.cif_id}"

    @property
    def cif_id(self):
        if cif_id := self.kwargs.get("cifname"):
            return cif_id
        else:
            return Path(self.path_cif).stem

    @property
    def cmap(self):
        return self.kwargs.get("cmap", None)

    @cmap.setter
    def cmap(self, cmap):
        if isinstance(cmap, (str, ListedColormap)) or cmap is None:
            self.kwargs["cmap"] = cmap
        else:
            raise TypeError(
                f"cmap must be str, ListedColormap, or None, not {type(cmap)}"
            )

    @property
    def num_patches(self):
        return self.kwargs.get("num_patches", (6, 6, 6))

    @property
    def figsize(self):
        return self.kwargs.get("figsize", DEFAULT_FIGSIZE)

    @figsize.setter
    def figsize(self, figsize):
        if not isinstance(figsize, Sequence):
            raise TypeError(f"figsize must be tuple or list, not {type(figsize)}")
        elif len(figsize) != 2:
            raise ValueError(f"figsize must be (float, float) not {figsize}")
        self.kwargs["figsize"] = figsize
        self._sync_cbar_kwargs()

    @property
    def view_init(self):
        return self.kwargs.get("view_init", DEFAULT_VIEW_INIT)

    @view_init.setter
    def view_init(self, view_init):
        if not isinstance(view_init, Sequence):
            raise TypeError(f"view_init must be tuple or list, not {type(view_init)}")
        elif len(view_init) != 2:
            raise ValueError(f"view_init must be (float, float) not {view_init}")
        self.kwargs["view_init"] = view_init

    @property
    def show_axis(self):
        return self.kwargs.get("show_colorbar", False)

    @show_axis.setter
    def show_axis(self, show_axis):
        if not isinstance(show_axis, bool):
            raise TypeError(f"show_axis must be bool, not {type(show_axis)}")
        self.kwargs["show_axis"] = show_axis

    @property
    def show_colorbar(self):
        return self.kwargs.get("show_colorbar", True)

    @show_colorbar.setter
    def show_colorbar(self, show_colorbar):
        if not isinstance(show_colorbar, bool):
            raise TypeError(f"show_colorbar must be bool, not {type(show_colorbar)}")
        self.kwargs["show_colorbar"] = show_colorbar

    @property
    def atomic_scale(self):
        return self.figsize[0] * self.figsize[1]

    def set_default(self):
        self.kwargs = {}
        self._sync_cbar_kwargs()

    def _sync_cbar_kwargs(self):
        figsize = self.figsize
        self.cbar_kwargs["labelpad"] = figsize[0] * 2
        self.cbar_kwargs["labelsize"] = figsize[0] * 1.5
        self.cbar_kwargs["fontsize"] = figsize[0]

    def _get_indice_inside_patch(self, patch_position, ep=0.0):
        def is_inside_patch(r_pos, pos_patch, num_patch):
            lower_bound = (pos_patch - ep) / num_patch < r_pos
            upper_bound = r_pos < (pos_patch + 1 + ep) / num_patch
            return np.logical_and(lower_bound, upper_bound)

        relative_position = self.s_atoms.get_scaled_positions()
        position_bools = [
            is_inside_patch(r_pos, pos_patch, num_patch)
            for r_pos, pos_patch, num_patch in zip(
                relative_position.T, patch_position, self.num_patches
            )
        ]  # x, y, z

        position_bool = np.logical_and.reduce(position_bools)  # total
        (indice,) = np.where(position_bool)
        return indice

    def _grid_attention_rank(self, rank):
        heatmap_grid = self.heatmap_grid
        sort = np.flip(
            np.unravel_index(np.argsort(heatmap_grid, axis=None), heatmap_grid.shape),
            axis=-1,
        )
        if isinstance(rank, int):
            return np.array(sort)[:, rank][np.newaxis, :]
        elif isinstance(rank, Iterable):
            return np.array(sort)[:, tuple(rank)].T
        else:
            raise TypeError(f"rank must be int or iterable, not {type(rank)}")

    def set_colorbar_options(self, default=False, **cbar_kwargs):
        if default:
            self.cbar_kwargs = get_default_cbar_kwargs(self.figsize)
        else:
            for key, value in cbar_kwargs.items():
                if key in self.cbar_kwargs:
                    self.cbar_kwargs[key] = value

    def draw_graph(
        self,
        minatt=0.000,
        maxatt=0.010,
        *,
        alpha=0.7,
        atomic_scale_factor=3,
        grid_scale_factor=1,
        att_scale_factor=2,
        return_fig=False,
        **kwargs,
    ):
        """
        Draw graph attention score figure in primitive unit cell
        :param minatt: (float) Minimum value of attention score (default : 0.000). A value smaller than minatt is treated as minatt.
        :param maxatt: (float) Maximum value of attention score (default : 0.010). A value larger than maxatt is treated as maxatt.
        :param alpha: (float) The alpha blending value, between 0 (transparent) and 1 (opaque).
        :param atomic_scale_factor: (float) The factors that determines atom size. (default = 3)
        :param grid_scale_factor: (float) The factors that determines grid size (default = 1)
        :param att_scale_factor: (float) The factor that determines attention-score overlay size (default = 2)
        :param return_fig : (bool) If True, matplotlib.figure.Figure and matplotlib.Axes3DSubplot are returned.
        :param kwargs:
            view_init : (float, float) view init from matplotlib
            show_axis : <bool> If True, axis are visible. (default : False)
            show_colorbar : <bool> If True, colorbar are visible. (default : False)
            cmap : (str or matplotlib.colors.ListedColormap) color map used in figure. (default : None)
        """
        heatmap_graph = self.heatmap_graph
        lattice = self.p_lattice
        atoms = self.p_atoms

        fig, ax = get_fig_ax(**self.kwargs)
        cmap = get_cmap(kwargs.get("cmap", self.cmap))
        set_fig_ax(ax, **kwargs)

        draw_cell(ax, lattice, color="black")
        draw_atoms(
            ax, atoms, self.atomic_scale * atomic_scale_factor * grid_scale_factor
        )

        colors = cmap(scaler(heatmap_graph, minatt, maxatt))
        atomic_scale = (
            self.atomic_scale
            * att_scale_factor
            * grid_scale_factor
            * atomic_scale_factor
        )
        draw_heatmap_graph(ax, atoms, colors, atomic_scale, alpha)

        if kwargs.get("show_colorbar", self.show_colorbar):
            draw_colorbar(fig, ax, cmap, minatt, maxatt, **self.cbar_kwargs)

        set_axes_equal(ax, scale_factor=grid_scale_factor)
        if return_fig:
            return fig, ax
        else:
            plt.show()


    def draw_image_1d(
        self,
        top_n=10,
        return_fig=False,
    ):
        heatmap_image_1d = self.heatmap_image_1d
        image_1d=self.image_1d
        max_birth=self.max_birth_1d
        max_persistence=self.max_persistence_1d

        fig,ax=plot_image(image_1d,max_birth,max_persistence,heatmap_image_1d,top_n)

        if return_fig:
            return fig, ax
        else:
            plt.show()

    def draw_image_2d(
        self,
        top_n=10,
        return_fig=False,
    ):
        heatmap_image_2d = self.heatmap_image_2d
        image_2d=self.image_2d
        max_birth=self.max_birth_2d
        max_persistence=self.max_persistence_2d

        fig,ax=plot_image(image_2d,max_birth,max_persistence,heatmap_image_2d,top_n)
        
        if return_fig:
            return fig, ax
        else:
            plt.show()

    # def draw_grid(
    #     self,
    #     minatt=0.000,
    #     maxatt=0.01,
    #     *,
    #     patch_list=None,
    #     remove_under_minatt=False,
    #     alpha=0.8,
    #     atomic_scale_factor=1,
    #     grid_scale_factor=1,
    #     return_fig=False,
    #     **kwargs,
    # ):
    #     """
    #     Draw grid attention score figure in supercell
    #     :param minatt: (float) Minimum value of attention score (default : 0.000). A value smaller than minatt is treated as minatt.
    #     :param maxatt: (float) Maximum value of attention score (default : 0.010). A value larger than maxatt is treated as maxatt.
    #     :param patch_list: (list) list of patch position that plot in figure. Draw all patches if not specified.
    #     :param remove_under_minatt: (bool) If True, do not draw a patch with an attention value lower than minatt.
    #     :param alpha: (float) The alpha blending value, between 0 (transparent) and 1 (opaque).
    #     :param atomic_scale_factor: (float) The factors that determines atom size. (default = 1)
    #     :param grid_scale_factor: (float) The factors that determines grid size (default = 3)
    #     :param return_fig : (bool) If True, matplotlib.figure.Figure and matplotlib.Axes3DSubplot are returned.
    #     :param kwargs:
    #         view_init : (float, float) view init from matplotlib
    #         show_axis : <bool> If True, axis are visible. (default : False)
    #         show_colorbar : <bool> If True, colorbar are visible. (default : False)
    #         cmap : (str or matplotlib.colors.ListedColormap) color map used in figure. (default : None)
    #     """
    #     heatmap_grid = self.heatmap_grid
    #     lattice = self.s_lattice
    #     atoms = self.s_atoms
    #     fig, ax = get_fig_ax(**self.kwargs)
    #     cmap = get_cmap(kwargs.get("cmap", self.cmap))
    #     set_fig_ax(ax, **kwargs)

    #     # patch_list
    #     if patch_list is None:
    #         patch_list = product(*[range(n) for n in self.num_patches])
    #     elif isinstance(patch_list, Iterable):
    #         pass
    #     else:
    #         raise TypeError(
    #             f"patch_list must be iterable object, not {type(patch_list)}"
    #         )

    #     # get position and heatmap
    #     positions = []
    #     sc_heatmap_grid = []
    #     for i, j, k in patch_list:
    #         att = heatmap_grid[i, j, k]
    #         if remove_under_minatt and att < minatt:
    #             continue
    #         positions.append((i, j, k))
    #         sc_heatmap_grid.append(scaler(att, minatt, maxatt))

    #     sc_heatmap_grid = np.array(sc_heatmap_grid)
    #     heatmap = cmap(sc_heatmap_grid.flatten())

    #     # draw
    #     draw_heatmap_grid(
    #         ax,
    #         positions,
    #         heatmap,
    #         lattice,
    #         num_patches=self.num_patches,
    #         alpha=alpha,
    #     )
    #     draw_cell(ax, lattice, color="black")
    #     draw_atoms(
    #         ax, atoms, self.atomic_scale * atomic_scale_factor * grid_scale_factor
    #     )

    #     if kwargs.get("show_colorbar", self.show_colorbar):
    #         draw_colorbar(fig, ax, cmap, minatt, maxatt, **self.cbar_kwargs)

    #     set_axes_equal(ax, scale_factor=grid_scale_factor)
    #     if return_fig:
    #         return fig, ax
    #     else:
    #         plt.show()

    # def draw_grid_with_attention_rank(
    #     self,
    #     rank,
    #     minatt=0.000,
    #     maxatt=0.010,
    #     *,
    #     remove_under_minatt=False,
    #     alpha=0.8,
    #     atomic_scale_factor=1,
    #     grid_scale_factor=1,
    #     return_fig=False,
    #     **kwargs,
    # ):
    #     """
    #     Draw grid attention score figure in supercell
    #     :param rank:  (int or iterable) The rank (int) or iterable of ranks (list, np.array, tuple, range, etc) of the patch you want to draw.
    #                     Rank means that the attention score is listed in the order of high.
    #     :param minatt: (float) Minimum value of attention score (default : 0.000). A value smaller than minatt is treated as minatt.
    #     :param maxatt: (float) Maximum value of attention score (default : 0.010). A value larger than maxatt is treated as maxatt.
    #     :param remove_under_minatt: (bool) If True, do not draw a patch with an attention value lower than minatt.
    #     :param alpha: (float) The alpha blending value, between 0 (transparent) and 1 (opaque).
    #     :param atomic_scale_factor: (float) The factors that determines atom size. (default = 1)
    #     :param grid_scale_factor: (float) The factors that determines grid size (default = 3)
    #     :param return_fig : (bool) If True, matplotlib.figure.Figure and matplotlib.Axes3DSubplot are returned.
    #     :param kwargs:
    #         view_init : (float, float) view init from matplotlib
    #         show_axis : <bool> If True, axis are visible. (default : False)
    #         show_colorbar : <bool> If True, colorbar are visible. (default : False)
    #         cmap : (str or matplotlib.colors.ListedColormap) color map used in figure. (default : None)
    #     """
    #     rank = self._grid_attention_rank(rank)
    #     return self.draw_grid(
    #         minatt,
    #         maxatt,
    #         patch_list=rank,
    #         remove_under_minatt=remove_under_minatt,
    #         alpha=alpha,
    #         atomic_scale_factor=atomic_scale_factor,
    #         grid_scale_factor=grid_scale_factor,
    #         return_fig=return_fig,
    #         **kwargs,
    #     )

    # def draw_specific_patch(
    #     self,
    #     patch_position,
    #     ep=0.5,
    #     *,
    #     color=True,
    #     alpha=0.5,
    #     minatt=0.000,
    #     maxatt=0.010,
    #     atomic_scale_factor=5,
    #     grid_scale_factor=1,
    #     return_fig=False,
    #     **kwargs,
    # ):
    #     """
    #     Draw one specific patch with neighbor atoms.
    #     :param patch_position:  (list) patch position that plot in figure.
    #     :param ep: (float) Distance of patches to be visualized around target patches (default = 0.5)
    #     :param color: (bool) If True, paint patch's color that indicate attention grid
    #     :param minatt: (float) Minimum value of attention score (default : 0.000). A value smaller than minatt is treated as minatt.
    #     :param maxatt: (float) Maximum value of attention score (default : 0.010). A value larger than maxatt is treated as maxatt.
    #     :param alpha: (float) The alpha blending value, between 0 (transparent) and 1 (opaque).
    #     :param atomic_scale_factor: (float) The factors that determines atom size. (default = 1)
    #     :param grid_scale_factor: (float) The factors that determines grid size (default = 3)
    #     :param return_fig : (bool) If True, plt.figure and plt.axes are returned.
    #     :param kwargs:
    #         view_init : (float, float) view init from matplotlib
    #         show_axis : <bool> If True, axis are visible. (default : False)
    #         show_colorbar : <bool> If True, colorbar are visible. (default : False)
    #         cmap : (str or matplotlib.colors.ListedColormap) color map used in figure. (default : None)
    #     """
    #     lattice = self.s_lattice
    #     atoms = self.s_atoms
    #     indice = self._get_indice_inside_patch(patch_position, ep=ep)
    #     s_point = np.sum(lattice.T / 6 * patch_position, axis=1)
    #     fig, ax = get_fig_ax(**self.kwargs)
    #     cmap = get_cmap(kwargs.get("cmap", self.cmap))
    #     set_fig_ax(ax, **kwargs)

    #     if ep > 0.5:
    #         scale_factor = (
    #             self.atomic_scale * atomic_scale_factor * grid_scale_factor / ep
    #         )
    #     else:
    #         scale_factor = (
    #             self.atomic_scale * atomic_scale_factor * grid_scale_factor * 2
    #         )

    #     draw_cell(ax, lattice / 6, s_point=s_point, color="black")
    #     draw_atoms(ax, atoms[indice], atomic_scale=scale_factor)

    #     if color:
    #         heatmap_grid = self.heatmap_grid
    #         patch_color = scaler(heatmap_grid[tuple(patch_position)], minatt, maxatt)
    #         patch_color = cmap(np.array([patch_color]))
    #         position = [patch_position]
    #         draw_heatmap_grid(
    #             ax,
    #             position,
    #             patch_color,
    #             lattice=lattice,
    #             num_patches=self.num_patches,
    #             alpha=alpha,
    #         )

    #     if kwargs.get("show_colorbar", self.show_colorbar):
    #         draw_colorbar(fig, ax, cmap, minatt, maxatt, **self.cbar_kwargs)

    #     set_axes_equal(ax, grid_scale_factor)

    #     if return_fig:
    #         return fig, ax
    #     else:
    #         plt.show()

    # def draw_specific_patch_with_attention_rank(
    #     self,
    #     rank,
    #     ep=0.5,
    #     *,
    #     color=True,
    #     alpha=0.5,
    #     minatt=0.000,
    #     maxatt=0.010,
    #     atomic_scale_factor=5,
    #     grid_scale_factor=1,
    #     return_fig=False,
    #     **kwargs,
    # ):
    #     """
    #     Draw one specific patch with neighbor atoms.
    #     :param rank : (int or iterable) The rank (int) of the patch you want to draw.
    #                     Rank means that the attention score is listed in the order of high.
    #     :param ep: (float) Distance of patches to be visualized around target patches (default = 0.5)
    #     :param color: (bool) If True, paint patch's color that indicate attention grid
    #     :param minatt: (float) Minimum value of attention score (default : 0.000). A value smaller than minatt is treated as minatt.
    #     :param maxatt: (float) Maximum value of attention score (default : 0.010). A value larger than maxatt is treated as maxatt.
    #     :param alpha: (float) The alpha blending value, between 0 (transparent) and 1 (opaque).
    #     :param atomic_scale_factor: (float) The factors that determines atom size. (default = 1)
    #     :param grid_scale_factor: (float) The factors that determines grid size (default = 3)
    #     :param return_fig : (bool) If True, matplotlib.figure.Figure and matplotlib.Axes3DSubplot are returned.
    #     :param kwargs:
    #         view_init : (float, float) view init from matplotlib
    #         show_axis : <bool> If True, axis are visible. (default : False)
    #         show_colorbar : <bool> If True, colorbar are visible. (default : False)
    #         cmap : (str or matplotlib.colors.ListedColormap) color map used in figure. (default : None)
    #     """
    #     if not isinstance(rank, int):
    #         raise TypeError(f"rank must be int, not {type(rank)}")

    #     rank = self._grid_attention_rank(rank).squeeze()
    #     return self.draw_specific_patch(
    #         rank,
    #         ep=ep,
    #         color=color,
    #         minatt=minatt,
    #         maxatt=maxatt,
    #         atomic_scale_factor=atomic_scale_factor,
    #         grid_scale_factor=grid_scale_factor,
    #         alpha=alpha,
    #         return_fig=return_fig,
    #         **kwargs,
    #     )

    # def animate(self, func, frame=360, interval=20, savefile=None, fps=30):
    #     def turn(i, ax, fig, **kwargs):
    #         view_init = kwargs.get("view_init", self.view_init)
    #         ax.view_init(elev=view_init[0], azim=i)
    #         return fig

    #     @wraps(func)
    #     def wrapper(*args, **kwargs):
    #         kwargs["return_fig"] = True
    #         fig, ax = func(*args, **kwargs)
    #         anim = animation.FuncAnimation(
    #             fig,
    #             partial(turn, ax=ax, fig=fig, **kwargs),
    #             init_func=lambda: fig,
    #             frames=frame,
    #             interval=interval,
    #             blit=True,
    #         )
    #         if savefile:
    #             anim.save(savefile, fps=fps, dpi=300)

    #         plt.show()
    #         return anim

    #     return wrapper



import functools
from typing import Optional
import torch
from torch.utils.data import DataLoader, random_split
from pytorch_lightning import LightningDataModule
from MCRT.data_processor.dataset import Dataset

class Datamodule(LightningDataModule):
    def __init__(self, _config):
        super().__init__()

        self.data_dir = _config["root_dataset"]
        self.max_num_nbr = _config["max_num_nbr"]
        self.mask_probability = _config["mask_probability"]
        self.num_ap=_config["num_ap"]
        self.n_dist=_config["n_dist"]
        self.n_angle=_config["n_angle"]
        self.max_graph_len = _config["max_graph_len"]
        self.angle_nbr = _config["angle_nbr"]
        self.num_workers = _config["num_workers"]
        self.batch_size = _config["per_gpu_batchsize"]
        self.eval_batch_size = self.batch_size
        self.downstream = _config["downstream"]
        self.nbr_fea_len = _config["nbr_fea_len"]
        self.if_conv = _config["if_conv"]
        self.pos_emb = _config["pos_emb"]
        self.if_alignn = _config["if_alignn"]
        self.if_image = _config["if_image"]
        self.if_grid = _config["if_grid"]
        self.read_from_pickle = _config["read_from_pickle"]
        self.tasks = [k for k, v in _config["loss_names"].items() if v > 0]

    @property
    def dataset_cls(self):
        return Dataset
    
    def set_train_dataset(self):
        self.train_dataset = self.dataset_cls(  
                data_dir=self.data_dir,
                split="train",
                max_num_nbr = self.max_num_nbr,
                nbr_fea_len=self.nbr_fea_len,
                num_ap=self.num_ap,
                n_dist=self.n_dist,
                n_angle=self.n_angle,
                downstream=self.downstream,
                tasks=self.tasks,
                if_conv=self.if_conv,
                pos_emb=self.pos_emb,
                if_alignn=self.if_alignn,
                if_image=self.if_image,
                if_grid=self.if_grid,
                read_from_pickle=self.read_from_pickle,
        )

    def set_val_dataset(self):
        self.val_dataset = self.dataset_cls(  
                data_dir=self.data_dir,
                split="val",
                max_num_nbr = self.max_num_nbr,
                nbr_fea_len=self.nbr_fea_len,
                num_ap=self.num_ap,
                n_dist=self.n_dist,
                n_angle=self.n_angle,
                downstream=self.downstream,
                tasks=self.tasks,
                if_conv=self.if_conv,
                pos_emb=self.pos_emb,
                if_alignn=self.if_alignn,
                if_image=self.if_image,
                if_grid=self.if_grid,
                read_from_pickle=self.read_from_pickle,
        )

    def set_test_dataset(self):
        self.test_dataset = self.dataset_cls(  
                data_dir=self.data_dir,
                split="test",
                max_num_nbr = self.max_num_nbr,
                nbr_fea_len=self.nbr_fea_len,
                num_ap=self.num_ap,
                n_dist=self.n_dist,
                n_angle=self.n_angle,
                downstream=self.downstream,
                tasks=self.tasks,
                if_conv=self.if_conv,
                pos_emb=self.pos_emb,
                if_alignn=self.if_alignn,
                if_image=self.if_image,
                if_grid=self.if_grid,
                read_from_pickle=self.read_from_pickle,
        )

    def setup(self, stage: Optional[str] = None):
        if stage in (None, "fit"):
            self.set_train_dataset()
            self.set_val_dataset()

        if stage in (None, "test"):
            self.set_test_dataset()


        self.collate = functools.partial(
            self.dataset_cls.collate,
            max_graph_len=self.max_graph_len,
            angle_nbr=self.angle_nbr,
            mask_probability = self.mask_probability,
            tasks=self.tasks,
            if_conv=self.if_conv,
            pos_emb=self.pos_emb,
            if_alignn=self.if_alignn,
        )

    def train_dataloader(self) -> DataLoader:
        return DataLoader(
            self.train_dataset,
            batch_size=self.batch_size,
            num_workers=self.num_workers,
            collate_fn=self.collate,
            shuffle=True,
            persistent_workers=True,
        )

    def val_dataloader(self) -> DataLoader:
        return DataLoader(
            self.val_dataset,
            batch_size=self.batch_size,
            num_workers=self.num_workers,
            collate_fn=self.collate,
            persistent_workers=True,
        )

    def test_dataloader(self) -> DataLoader:
        return DataLoader(
            self.test_dataset,
            batch_size=self.batch_size,
            num_workers=self.num_workers,
            collate_fn=self.collate,
        )



import copy
import math
import ase.io
from ase.build import make_supercell
from itertools import product
from functools import lru_cache
from pathlib import Path
from collections.abc import Iterable
import numpy as np
import os
import pickle
from mpl_toolkits.mplot3d.art3d import Poly3DCollection
from pymatgen.io.cif import CifParser
from pymatgen.io.ase import AseAtomsAdaptor
import torch
import pytorch_lightning as pl
from MCRT.modules.McrtModule import MCRT_Module
from MCRT.modules.datamodule import Datamodule
from MCRT.config import config
import csv

@lru_cache
def get_model_and_datamodule(model_path, data_root, downstream=""):
    _config = config()
    _config["visualize"] = True
    _config["per_gpu_batchsize"] = 1
    _config["data_root"] = data_root
    _config["root_dataset"] = data_root
    _config["load_path"] = model_path
    _config["test_only"] = True
    _config["log_dir"] = "result_visualization"
    _config["downstream"] = downstream
    _config["num_workers"] = 0

    pl.seed_everything(_config["seed"])
    model = MCRT_Module(_config)
    model.setup("test")
    model.eval()
    model.to("cpu")

    dm = Datamodule(_config)
    dm.setup("test")
    data_iter = dm.test_dataloader()

    return model, data_iter


@lru_cache
def get_batch_from_index(data_iter, batch_id):
    iter_ = iter(data_iter)
    for _ in range(batch_id):
        next(iter_)
    batch = next(iter_)
    return batch


@lru_cache
def get_batch_from_cif_id(data_iter, cif_id):
    cif_id = Path(cif_id).stem
    iter_ = iter(data_iter)
    while True:
        try:
            batch = next(iter_)
        except StopIteration:
            raise ValueError(f"There are no {cif_id} in dataset")
        else:
            batch_id = batch["cif_id"][0]
            if batch_id == cif_id:
                return batch


@lru_cache
def get_primitive_structure(path_cif, tolerance=2.0):
    (st,) = CifParser(path_cif, occupancy_tolerance=tolerance).get_structures(
        primitive=True
    )
    return st

@lru_cache
def get_structure(
    path_cif, make_supercell=False, dtype="ase", *, max_length=60, min_length=15
):
    """
    get primitive structure from path_cif
    :param path_cif: <str> path for cif file
    :param make_supercell: <bool> if True,
    :param dtype: <str> -> ['ase', 'pymatgen'] return type for structure.
    :param max_length: <int/float> max p_lattice length of structure file (Å)
    :param min_length: <int/float> min p_lattice length of structure file (Å)
    :return: <pymatgen.Structure> structure file from path cif
    """
    def transform_path_cif_to_pickle(path_cif):
        directory, filename = os.path.split(path_cif)
        new_filename = os.path.splitext(filename)[0] + ".pickle"
        parent_directory = os.path.dirname(directory)
        new_directory = os.path.join(parent_directory, "pickles")
        new_path = os.path.join(new_directory, new_filename)
        return new_path
    file_structure = transform_path_cif_to_pickle(path_cif)

    with open(file_structure, "rb") as f:
        crystal_data = pickle.load(f)
    structure=crystal_data["structure"]

    # structure.to(fmt="cif", filename="MCRT/visualize/results/output.cif")

    adaptor = AseAtomsAdaptor()
    atoms = adaptor.get_atoms(structure)

    if make_supercell:
        atoms = get_supercell_structure(atoms, max_length, min_length)
    # else:
    #     atoms = get_supercell_structure(atoms, max_length, 4)

    if dtype == "pymatgen":
        return AseAtomsAdaptor().get_structure(atoms)
    elif dtype == "ase":
        return atoms
    else:
        raise TypeError(f"type must be ase or pymatgen, not {dtype}")


def get_supercell_structure(atoms, max_length=60, min_length=15):
    """
    get supercell atoms from <ase.Atoms>
    :param atoms: <ase.Atoms> object
    :param max_length: <int/float> max p_lattice length of structure file (Å)
    :param min_length: <int/float> min p_lattice length of structure file (Å)
    :return: <ase.Atoms or pymatgen.Structure> structure type.
    """
    scale_abc = []
    for l in atoms.cell.cellpar()[:3]:
        if l > max_length:
            raise ValueError(
                f"primitive p_lattice is larger than max_length {max_length}"
            )
        elif l < min_length:
            scale_abc.append(math.ceil(min_length / l))
        else:
            scale_abc.append(1)

    m = np.zeros([3, 3])
    np.fill_diagonal(m, scale_abc)
    atoms = make_supercell(atoms, m)
    return atoms


def cuboid_data(position, color=None, num_patches=(6, 6, 6), lattice=None):
    """
    Get cuboid plain data from position and size data
    :param position: <list/tuple> patch positions => [x, y, z]
    :param color: <list/tuple> colors => [r, g, b, w]
    :param num_patches: number of patches in each axis (default : (6, 6, 6))
    :param lattice: <np.ndarray> p_lattice vector for unit p_lattice
    :return: <tuple> (list of plain vector, list of color vector)
    """
    if isinstance(num_patches, (tuple, list)):
        num_patches = np.array(num_patches)
    elif not isinstance(num_patches, np.ndarray):
        raise TypeError(f"num_patches must be tuple or list, not {type(num_patches)}")

    bound = np.array([[0, 1] for _ in range(3)]) + np.array(position)[:, np.newaxis]
    vertex = np.array(list(product(*bound)))
    plane_ls = []
    for i, (dn, up) in enumerate(bound):
        plane1 = np.matmul(vertex[vertex[:, i] == dn], lattice / num_patches)
        plane2 = np.matmul(vertex[vertex[:, i] == up], lattice / num_patches)

        plane_ls.append(plane1)
        plane_ls.append(plane2)

    plane_ls = np.array(plane_ls).astype("float")
    plane_ls[:, [0, 1], :] = plane_ls[:, [1, 0], :]

    color_ls = np.repeat(color[np.newaxis, :], 6, axis=0)

    return plane_ls, color_ls


def plot_cube(positions, colors, lattice, num_patches=(6, 6, 6), **kwargs):
    """
    help function for draw 3d cube plot
    :param positions: <list> list of patch position
    :param colors: <list -> list> list of color codes [r, g, b, w]
    :param lattice: <np.ndarray> p_lattice vector for unit p_lattice
    :param num_patches: number of patches in each axis (default : (6, 6, 6))
    :param kwargs: kwargs for <matplotlib.Poly3DCollection>
    :return: <matplotlib.Poly3DCollection> cuboid matplotlib object
    """

    data = [
        cuboid_data(pos, color, num_patches=num_patches, lattice=lattice)
        for pos, color in zip(positions, colors)
    ]
    plain_ls, color_ls = zip(*data)

    return Poly3DCollection(
        np.concatenate(plain_ls), facecolors=np.concatenate(color_ls), **kwargs
    )


def get_heatmap(out, batch_idx, graph_len=1023,image_len=100, skip_cls=True):
    """
    attention rollout  in "Quantifying Attention Flow in Transformers" paper.
    :param out: output of model.infer(batch)
    :param batch_idx: batch index
    :param graph_len: the length of grid embedding
    :param skip_cls: <bool> If True, class token is ignored.
    :return: <np.ndarray> heatmap graph, heatmap grid
    """
    print("start cal heatmap")
    attn_weights = torch.stack(
        out["attn_weights"]
    )  # [num_layers, B, num_heads, max_len, max_len]
    att_mat = attn_weights[:, batch_idx]  # [num_layers, num_heads, max_len, max_len]

    image=out["image"]# list 2 [B,1,h,w]
    max_birth_1d=out["max_birth_1d"]#list [B]
    max_persistence_1d=out["max_persistence_1d"]
    max_birth_2d=out["max_birth_2d"]
    max_persistence_2d=out["max_persistence_2d"]
    # transform into numpy
    image_1d=image[0][batch_idx, 0, :, :].detach().numpy() #[h,w]
    image_2d=image[1][batch_idx, 0, :, :].detach().numpy() #[h,w]
    max_birth_1d=max_birth_1d[batch_idx].detach().numpy()
    max_persistence_1d=max_persistence_1d[batch_idx].detach().numpy()
    max_birth_2d=max_birth_2d[batch_idx].detach().numpy()
    max_persistence_2d=max_persistence_2d[batch_idx].detach().numpy()

    # Average the attention weights across all heads.
    att_mat = torch.mean(att_mat, dim=1)  # [num_layers, max_len, max_len]

    # To account for residual connections, we add an identity matrix to the
    # attention matrix and re-normalize the weights.
    residual_att = torch.eye(att_mat.size(1))
    aug_att_mat = att_mat + residual_att

    aug_att_mat = aug_att_mat / aug_att_mat.sum(dim=-1).unsqueeze(
        -1
    )  # [num_layers, max_len, max_len]
    aug_att_mat = aug_att_mat.detach().numpy()  # prevent from memory leakage

    # Recursively multiply the weight matrices
    joint_attentions = np.zeros(aug_att_mat.shape)  # [num_layers, max_len, max_len]
    joint_attentions[0] = aug_att_mat[0]

    for n in range(1, aug_att_mat.shape[0]):
        joint_attentions[n] = np.matmul(aug_att_mat[n], joint_attentions[n - 1])

    v = joint_attentions[-1]  # [max_len, max_len]

    # Don't drop class token when normalizing
    if skip_cls:
        v_ = v[0][1:]  # skip cls token
        cost_graph = v_[:graph_len]  # / v_.max()
        cost_image = v_[graph_len:]  # / v_.max()
        heatmap_graph = cost_graph
        heatmap_image_1d = cost_image[1:image_len+3]  # omit sep tokens
        heatmap_image_2d = cost_image[image_len+3:]  # omit scale tokens
    else:
        v_ = v[0]
        cost_graph = v_[: graph_len + 1]  # / v_.max()
        cost_image = v_[graph_len + 1 :]  # / v_.max()
        heatmap_graph = cost_graph[1:]  # omit cls token
        heatmap_image_1d = cost_image[1:image_len+3] # omit sep tokens
        heatmap_image_2d = cost_image[image_len+3:] # omit scale tokens

    return heatmap_graph, heatmap_image_1d, heatmap_image_2d,image_1d,image_2d,max_birth_1d,max_persistence_1d,max_birth_2d,max_persistence_2d,


def scaler(value, min_att, max_att):
    if isinstance(value, float):
        if value > max_att:
            value = max_att
        elif value < min_att:
            value = min_att
        return float((value - min_att) / (max_att - min_att))

    elif isinstance(value, np.ndarray):
        value = copy.deepcopy(value)
        value[value > max_att] = max_att
        value[value < min_att] = min_att
        return (value - min_att) / (max_att - min_att)
    elif isinstance(value, Iterable):
        return scaler(np.array(list(value), dtype="float"), min_att, max_att)
    else:
        raise TypeError(f"value must be float, list, or np.array, not {type(value)}")



import warnings
warnings.filterwarnings("ignore")
import sys
import os
import json
import random
import numpy as np
import torch
import pickle
import csv
import dgl
from pymatgen.core.structure import Structure
from pymatgen.symmetry.analyzer import SpacegroupAnalyzer
from MCRT.data_processor.cif_to_dataset import (
    jarvis_atoms_to_dgl_graph,
    compute_bond_cosines,
    convert_structures_to_jarvis_atoms,
)
'''
This file is used for constructing the dataset for pytorch
'''  

class Dataset(torch.utils.data.Dataset):
    def __init__(
        self,
        data_dir: str,
        split: str,
        max_num_nbr: int=12,
        nbr_fea_len: int=41,
        num_ap:int=200,
        n_dist:int=100,
        n_angle:int=100,
        downstream="",
        tasks=[],
        if_conv=False,
        pos_emb='relative',
        if_alignn = True,
        if_image = True,
        if_grid = False,
        read_from_pickle = True,
    ):
        """
        Dataset for pretrained MCRT.
        Args:
            data_dir (str): where dataset cif files
        """
        super().__init__()
        self.data_dir = data_dir
        assert split in {"train", "test", "val"}
        self.split = split
        self.max_num_nbr = max_num_nbr
        self.nbr_fea_len = nbr_fea_len
        self.num_ap=num_ap
        self.n_dist=n_dist
        self.n_angle=n_angle
        self.downstream=downstream
        self.tasks=tasks
        self.if_conv=if_conv
        self.pos_emb=pos_emb
        self.if_alignn=if_alignn
        self.if_image = if_image
        self.if_grid = if_grid
        self.read_from_pickle = read_from_pickle

        if "cdp" in self.tasks:
            density_file = os.path.join(self.data_dir, 'density.csv')
            self.id_density_data=self.get_prop(density_file)
        if "sgp" in self.tasks:
            space_group_file = os.path.join(self.data_dir, 'space_group.csv')
            self.id_space_group_data=self.get_prop(space_group_file)
        if "sep" in self.tasks: 
            space_group_file = os.path.join(self.data_dir, 'space_group.csv')
            self.id_space_group_data=self.get_prop(space_group_file)
            symm_elem_file = os.path.join(self.data_dir, 'simplyfied_symm_elem.json')
            with open(symm_elem_file, 'r') as f:
                data_dict = json.load(f)
            self.symm_elem_data=data_dict

        self.cif_path=os.path.join(self.data_dir,"cifs")

        if self.read_from_pickle:
            self.pickle_path=os.path.join(self.data_dir,"pickles") # pickles include the info for atom_label (for "map") and atm_label (for "apc")
            if not os.path.exists(self.pickle_path):
                print('!!!pickle path does not exist, generate graph input for finetune in real time. NOTE:it doesnot support pretrain because label for pretrain is time-consuming!!!')
                self.read_from_pickle=False
                assert os.path.exists(self.cif_path), 'cifs does not exist!'

        if if_image:
            self.image_path=os.path.join(self.data_dir,"imgs") 
            assert os.path.exists(self.image_path), 'images does not exist!'

        if if_grid:
            self.grid_path=os.path.join(self.data_dir,"grids") 
            assert os.path.exists(self.grid_path), 'grids does not exist!'

        dataset_split = self.load_dataset_split()


        conditions = [] # restore all cif conditions
        if self.read_from_pickle:
            pickle_cif_ids = set(os.path.splitext(name)[0] for name in os.listdir(self.pickle_path) if name.endswith('.pickle'))
            conditions.append(pickle_cif_ids)

        if if_image:
            image_cif_ids = set(os.path.splitext(name)[0] for name in os.listdir(self.image_path) if name.endswith('.npy'))
            conditions.append(image_cif_ids)

        if if_grid:
            grid_cif_ids = set(os.path.splitext(name)[0] for name in os.listdir(self.grid_path) if name.endswith('.grid'))
            griddata16_cif_ids = set(os.path.splitext(name)[0] for name in os.listdir(self.grid_path) if name.endswith('.griddata16'))
            combined_grid_ids = grid_cif_ids & griddata16_cif_ids
            conditions.append(combined_grid_ids)

        # Combine all conditions
        if conditions:
            combined_conditions = set.intersection(*conditions)
            self.cif_ids = [cid for cid in dataset_split[self.split] if cid in combined_conditions]
        else:
            self.cif_ids = dataset_split[self.split]

        if self.downstream:
            downstream_file = os.path.join(self.data_dir, f"{downstream}.csv")
            self.id_downstream_data=self.get_prop(downstream_file)
            downstream_cif_keys = set(self.id_downstream_data.keys())
            self.cif_ids = [cid for cid in self.cif_ids if cid in downstream_cif_keys]

    def load_dataset_split(self):
        # json_path = os.path.join(self.data_dir, 'dataset_split.json')
        
        json_path = os.path.join(self.data_dir, 'dataset_split.json')
        with open(json_path, 'r') as file:
            return json.load(file)

    def get_prop(self, prop_file):
        assert os.path.exists(prop_file), f'{prop_file} does not exist!'
        print(f"reading {prop_file}")
        id_prop_data ={}
        with open(prop_file) as f:
            reader = csv.reader(f)
            for row in reader:
                key = row[0]
                try:
                    value = row[1]
                except ValueError:
                    continue
                id_prop_data[key] = value
        return id_prop_data

    def __len__(self):
        return len(self.cif_ids)

    @staticmethod
    def apply_mask_to_atom_num(atom_num, mask_probability=0.15):
        """
        Applies masking logic directly to atom numbers in the dataset.

        Args:
        atom_num (torch.Tensor): Tensor of atom types/numbers, labels for each atom.
        mask_probability (float): Probability of applying mask.

        Returns:
        masked_atom_num (torch.Tensor): Tensor of atom numbers after applying mask.
        atom_label (torch.Tensor): Labels tensor for loss computation, with -1 indicating atoms to ignore.
        """
        atom_label = torch.full_like(atom_num, -1)
        mask = torch.rand(atom_num.size(0)) < mask_probability
        masked_atom_num = atom_num.clone()

        for i in range(masked_atom_num.size(0)):
            if mask[i]:
                atom_label[i] = atom_num[i]# Set the correct label for masked atom
                decision = torch.rand(1).item()
                if decision < 0.8:  # 80% chance to replace with [MASK], in this case 0
                    masked_atom_num[i] = 0
                elif decision < 0.9:  # 10% chance for random replacement
                    masked_atom_num[i] = torch.randint(1, 119, (1,)).item()
                # Remaining 10% chance to keep unchanged
                
        return masked_atom_num, atom_label  

    def select_atom_pairs(self, atm_list, m):
        """    
        m: number of atom pairs from each crystal
        half of m atom pairs are from same molecule, remaining from different molecules,
        if possible number of atom pairs < m, number of selected pairs = max_pairs*2, max_pairs = min(m // 2, same_molecule_pairs_count, diff_molecule_pairs_count)
        """
        atom_pairs = torch.full((m, 2), -1, dtype=torch.long)  # Use -1 for invalid pairs
        ap_labels = torch.full((m,), -1, dtype=torch.long)  # Use -1 for invalid ap_labels

        
        # Collect information about molecules in the crystal
        molecule_atoms = {}
        for idx, molecule_id in enumerate(atm_list):
            molecule_atoms.setdefault(molecule_id, []).append(idx)

        # Calculate the number of possible same molecule pairs
        same_molecule_pairs_count = sum(len(atoms) * (len(atoms) - 1) // 2 for atoms in molecule_atoms.values())
        # Calculate the number of possible diff molecule pairs
        diff_molecule_pairs_count = 0
        molecule_ids = list(molecule_atoms.keys())
        for idx1, mol1 in enumerate(molecule_ids):
                for mol2 in molecule_ids[idx1 + 1:]:
                    diff_molecule_pairs_count += len(molecule_atoms[mol1]) * len(molecule_atoms[mol2])

        max_pairs = min(m // 2, same_molecule_pairs_count, diff_molecule_pairs_count)

        selected_pairs = []
        selected_pairs_set = set()

            # Randomly select same molecule pairs
        while len(selected_pairs) < max_pairs:
                mol = random.choice(list(molecule_atoms.keys()))
                if len(molecule_atoms[mol]) > 1:
                    atom1, atom2 = random.sample(molecule_atoms[mol], 2)
                    pair = tuple(sorted((atom1, atom2)))  # Ensure the pair is sorted
                    if pair not in selected_pairs_set:
                        selected_pairs.append(pair)
                        ap_labels[len(selected_pairs) - 1] = 1  # Same molecule
                        selected_pairs_set.add(pair)

            # Select different molecule pairs
        while len(selected_pairs) < 2 * max_pairs:
                mol1, mol2 = random.sample(list(molecule_atoms.keys()), 2)
                atom1, atom2 = random.choice(molecule_atoms[mol1]), random.choice(molecule_atoms[mol2])
                pair = tuple(sorted((atom1, atom2)))  # Ensure the pair is sorted
                if pair not in selected_pairs_set:
                    selected_pairs.append(pair)
                    ap_labels[len(selected_pairs) - 1] = 0  # Different molecules
                    selected_pairs_set.add(pair)

            # atom_pairs[:len(selected_pairs)] = torch.tensor(selected_pairs)
        if len(selected_pairs) > 0:  # Ensure there are selected pairs
                    selected_pairs_tensor = torch.tensor(selected_pairs, dtype=torch.long)
                    atom_pairs[:selected_pairs_tensor.size(0)] = selected_pairs_tensor

        return atom_pairs, ap_labels #  [m, 2],[m] 
    
    def select_dist_pairs(self, distance_matrix, n_dist=100):
        dist_pairs = torch.full((n_dist, 2), -1, dtype=torch.long)
        dist_labels = torch.full((n_dist,), -1.0, dtype=torch.float)
        n_atoms = distance_matrix.shape[0]
        # possible pairs
        max_possible_pairs = n_atoms * (n_atoms - 1) // 2
        max_n_dist = min(n_dist, max_possible_pairs)


        dist_pairs_set = set()
        pair_count = 0

        while pair_count < max_n_dist:
            i = random.randint(0, n_atoms - 1)
            j = random.randint(0, n_atoms - 1)
            if i != j:
                pair = tuple(sorted((i, j)))
                if pair not in dist_pairs_set:
                    dist_pairs[pair_count] = torch.tensor(pair, dtype=torch.long)
                    dist_labels[pair_count] = distance_matrix[i, j]
                    dist_pairs_set.add(pair)
                    pair_count += 1
        return dist_pairs, dist_labels #  [n_dist, 2],[n_dist] 
    
    def select_angle_pairs(self, structure, n_angle=100):
        angle_pairs = torch.full((n_angle, 3), -1, dtype=torch.long)
        angle_labels = torch.full((n_angle,), -100.0, dtype=torch.float)
        n_atoms = len(structure)
        # possible triples
        max_possible_triples = n_atoms * (n_atoms - 1) * (n_atoms - 2) // 6
        max_n_angle = min(n_angle, max_possible_triples)

        angle_pairs_set = set()
        pair_count = 0

        while pair_count < max_n_angle:
            atoms = random.sample(range(n_atoms), 3)
            triple = tuple(atoms)
            if triple not in angle_pairs_set:
                # Calculate the angle
                vec1 = structure[atoms[1]].coords - structure[atoms[0]].coords
                vec2 = structure[atoms[1]].coords - structure[atoms[2]].coords
                cos_angle = self.calculate_cosine_angle(vec1, vec2)

                angle_pairs[pair_count] = torch.tensor(triple, dtype=torch.long)
                angle_labels[pair_count] = cos_angle
                angle_pairs_set.add(triple)
                pair_count += 1

        return angle_pairs, angle_labels

    @staticmethod
    def get_atom_graph(structure,radius=8,max_num_nbr=12):
        # RBF expand
        # gdf = GaussianDistance(dmin=0, dmax=8, step=0.2)
        all_nbrs = structure.get_all_neighbors(radius, include_index=True)
        all_nbrs = [sorted(nbrs, key=lambda x: x[1]) for nbrs in all_nbrs]
        nbr_fea_idx, nbr_fea = [], []
        for nbr in all_nbrs:
            if len(nbr) < max_num_nbr:
                nbr_fea_idx.append(list(map(lambda x: x[2], nbr)) +
                                    [0] * (max_num_nbr - len(nbr)))
                nbr_fea.append(list(map(lambda x: x[1], nbr)) +
                                [radius + 1.] * (max_num_nbr -
                                                        len(nbr)))
            else:
                nbr_fea_idx.append(list(map(lambda x: x[2],
                                                nbr[:max_num_nbr])))
                nbr_fea.append(list(map(lambda x: x[1],
                                            nbr[:max_num_nbr])))
        nbr_fea_idx, nbr_fea = np.array(nbr_fea_idx), np.array(nbr_fea)
        # nbr_fea = gdf.expand(nbr_fea)
        nbr_fea = torch.Tensor(nbr_fea) # it's actually distance now
        nbr_fea_idx = torch.LongTensor(nbr_fea_idx)
            
        return nbr_fea, nbr_fea_idx
    
    @staticmethod
    def get_gaussian_distance(distances, num_step:int=41, dmax=8, dmin=0, var=None):
        """
        Expands the distance by Gaussian basis, adapted from (https://github.com/txie-93/cgcnn.git)
        """

        assert dmin < dmax
        _filter = np.linspace(
            dmin, dmax, num_step
        ) 
        if var is None:
            var = (dmax-dmin)/(num_step-1)
        return np.exp(-((distances[..., np.newaxis] - _filter) ** 2) / var**2)
    
    
    @staticmethod
    def shuffle_structure(structure):
        # Create a list of indices and shuffle it
        indices = list(range(len(structure)))
        random.shuffle(indices)
        new_species = [structure[i].specie for i in indices]
        new_coords = [structure[i].frac_coords for i in indices]
        shuffled_structure = Structure(structure.lattice, new_species, new_coords)
        return shuffled_structure, indices
    
    @staticmethod
    def calculate_cosine_angle(v1, v2):
        cosine_angle = np.dot(v1, v2) / (np.linalg.norm(v1) * np.linalg.norm(v2))
        cosine_angle = np.clip(cosine_angle, -1, 1) #adjust wrong values due to limited precision
        return cosine_angle

    # def get_nearest_neighbors_cos_angles(self, structure, n_nbr=8):
    #     cos_angles_list = []
    #     for i, site in enumerate(structure):
    #         neighbors = structure.get_neighbors(site, r=max(structure.lattice.a,structure.lattice.b,structure.lattice.c)/2, include_index=True)
    #         neighbors.sort(key=lambda x: x[1])
    #         nearest_neighbors = neighbors[:n_nbr]
    #         cos_angles = []
    #         total_expected_angles = sum(range(n_nbr))
    #         # Calculate actual angles
    #         for j in range(n_nbr - 1):  # This ensures up to 7 loops for 8 neighbors
    #             actual_angles_this_round = 0
    #             for k in range(j + 1, len(nearest_neighbors)):
    #                 vec1 = nearest_neighbors[j][0].coords - site.coords
    #                 vec2 = nearest_neighbors[k][0].coords - site.coords
    #                 cos_angle = self.calculate_cosine_angle(vec1, vec2)
    #                 cos_angles.append(cos_angle)
    #                 actual_angles_this_round += 1
                
    #             # Dynamic padding for missing angles in this round
    #             expected_angles_this_round = n_nbr - 1 - j
    #             missing_angles_this_round = expected_angles_this_round - actual_angles_this_round
    #             cos_angles.extend([-100] * missing_angles_this_round)

    #         total_actual_angles = len(cos_angles)
    #         if total_actual_angles < total_expected_angles:
    #             cos_angles.extend([-100] * (total_expected_angles - total_actual_angles))
    #         cos_angles_list.append(cos_angles)

    #     return cos_angles_list

    def shuffle_structure_and_get_graph(self,crystal_data):
        """
        shuffle structures, generate atom graphs in real time.
        dimensions:
        atom_num:[n,1]; nbr_fea:[n,num_nbr,41]; nbr_fea_idx:[n,12]; distance_matrix:[n,n]; atom_to_molecule_list: [n,1]; 
        angles_list:[n,n_nbr*(n_nbr-1)/2].
        """
        structure=crystal_data["structure"]
        cif_id = crystal_data["cif_id"]
        original_atom_num = crystal_data["atom_num"]

        # # if if_alignn, do not shuffle
        # if not self.if_alignn:
        #     shuffled_structure, indices = self.shuffle_structure(structure)
        #     graph=None
        #     line_graph=None
        # else:
        #     shuffled_structure=structure
        #     indices=list(range(len(structure)))
        #     graph=crystal_data["graph"]
        #     line_graph=crystal_data["line_graph"]

        # only when train, shuffle the structure
        if self.split=="train":
            shuffled_structure, indices = self.shuffle_structure(structure)
        else:
            shuffled_structure=structure
            indices=list(range(len(structure)))

        if not self.if_alignn:
            graph=None
            line_graph=None
        else:
            atoms = convert_structures_to_jarvis_atoms(shuffled_structure)
            graph = jarvis_atoms_to_dgl_graph(
                        atoms,
                        "k-nearest",
                        8.0,
                        12,
                        True,
                    )
            graph.apply_edges(
                        lambda edges: {"distance": torch.norm(edges.data["coord_diff"], dim=1)}
                    )
            line_graph = graph.line_graph(shared=True)
            line_graph.apply_edges(compute_bond_cosines)

        # re-order atom_num
        atom_num = [original_atom_num[i] for i in indices]
        atom_num = torch.LongTensor(atom_num)

        #atom graph for cgcnn
        if self.if_conv:
            nbr_fea, nbr_fea_idx=self.get_atom_graph(structure=shuffled_structure,radius=8,max_num_nbr=self.max_num_nbr)
        else:
            nbr_fea=None
            nbr_fea_idx=None

        #distance_matrix
        distance_matrix=shuffled_structure.distance_matrix
        distance_matrix = torch.Tensor(distance_matrix)

        # positional embedding
        if self.pos_emb=='relative':
            # #bond angle list
            # angles_list=self.get_nearest_neighbors_cos_angles(structure=shuffled_structure, n_nbr=8)#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!for code test only
            original_angles_list = crystal_data["angles_list"]
            angles_list = [original_angles_list[i] for i in indices]
            angles_list = torch.Tensor(angles_list)
            abs_pos=None
        elif self.pos_emb=='absolute':
            abs_pos=shuffled_structure.cart_coords
            abs_pos=torch.Tensor(abs_pos)
            angles_list=None
        elif self.pos_emb=='both':
            # #bond angle list
            # angles_list=self.get_nearest_neighbors_cos_angles(structure=shuffled_structure, n_nbr=8)#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!for code test only
            original_angles_list = crystal_data["angles_list"]
            angles_list = [original_angles_list[i] for i in indices]
            angles_list = torch.Tensor(angles_list)
            abs_pos=shuffled_structure.cart_coords
            abs_pos=torch.Tensor(abs_pos)
        else:
            angles_list=None
            abs_pos=None
            pass

        # # mask atoms
        # if "map" in self.tasks:
        #     atom_num, atom_label = self.apply_mask_to_atom_num(atom_num, mask_probability=self.mask_probability)
        # else:
        #     atom_label = torch.full_like(atom_num, -1)

        # randomly select atom pairs and generate labels using atm_list
        if "apc" in self.tasks:
            original_atm_list = crystal_data["atm_list"]
            atm_list = [original_atm_list[i] for i in indices]
            atom_pairs, ap_labels = self.select_atom_pairs(atm_list, self.num_ap) #  [m, 2],[m]
        else:        
            atom_pairs = None
            ap_labels = None

        if "adp" in self.tasks:
            dist_pairs, dist_labels = self.select_dist_pairs(distance_matrix, self.n_dist)
        else:
            dist_pairs = None
            dist_labels = None

        if "aap" in self.tasks:
            angle_pairs, angle_labels = self.select_angle_pairs(shuffled_structure, self.n_angle)
        else:
            angle_pairs = None
            angle_labels = None

        if "ucp" in self.tasks:
            lattice = shuffled_structure.lattice
            lengths = np.array(lattice.lengths)
            angles = np.array(lattice.angles) 

            length_norm = 60.0  # normalize
            angle_norm = 180.0  # normalize
            norm_lengths = np.round(lengths / length_norm, 4)
            norm_angles = np.round(angles / angle_norm, 4)
            
            ucp_labels = list(norm_lengths) + list(norm_angles)  # [6]
        else:
            ucp_labels = None

        crystal_data={
                    "atom_num":atom_num,
                    "nbr_fea":nbr_fea, 
                    "nbr_fea_idx":nbr_fea_idx,
                    "distance_matrix":distance_matrix,
                    # "atm_list":atm_list,
                    "atom_pairs":atom_pairs,
                    "ap_labels":ap_labels,
                    "dist_pairs":dist_pairs,
                    "dist_labels":dist_labels,
                    "angle_pairs":angle_pairs,
                    "angle_labels":angle_labels,
                    "ucp_labels":ucp_labels,
                    "angles_list":angles_list,
                    "abs_pos":abs_pos,
                    "graph":graph,
                    "line_graph":line_graph,
                    "cif_id":cif_id,
                }
        return crystal_data

    @staticmethod
    def read_structure(file_path):
        try:
            structure = Structure.from_file(file_path, occupancy_tolerance=100.0)
            return structure
        except Exception as e:
            print(f"Error reading {file_path}: {e}", file=sys.stderr)

    @staticmethod
    def symmetry_operation(structure):
        # Redirect stdout and stderr to /dev/null
        with open(os.devnull, 'w') as devnull:
            os.dup2(devnull.fileno(), sys.stdout.fileno())
            os.dup2(devnull.fileno(), sys.stderr.fileno())
        try:
            analyzer = SpacegroupAnalyzer(structure)
            symmetrized_structure = analyzer.get_symmetrized_structure()
            return symmetrized_structure
        except Exception as e:
            print(f"Error during symmetry operation: {e}", file=sys.stderr)
    

    # pseudo code for angle matrix
    # function get_nearest_neighbors_cos_angles(structure, n_nbr, distance_threshold):
    #     initialize empty list cos_angles_list

    #     for each atom i in structure:
    #         # Step 1: Sort all neighbors by distance
    #         all_indices_sorted = sort_indices_by_distance(distance_matrix[i])
    #         all_distances_sorted = distances_of_sorted_indices(distance_matrix[i], all_indices_sorted)

    #         # Exclude the current atom itself from the neighbors
    #         remove_self(all_indices_sorted, all_distances_sorted)

    #         # Step 2: Get the distance of the nth nearest neighbor
    #         n_nbr_distance = all_distances_sorted[n_nbr - 1]

    #         # Step 3: Extend neighbors to include all atoms within the threshold distance from nth neighbor
    #         extended_count = n_nbr
    #         while extended_count < len(all_distances_sorted) and |all_distances_sorted[extended_count] - n_nbr_distance| < distance_threshold:
    #             extended_count += 1

    #         # Select the extended list of nearest neighbors
    #         nearest_neighbors_indices = all_indices_sorted[:extended_count]
    #         nearest_distances = all_distances_sorted[:extended_count]

    #         # Step 4: Group neighbors by similar distances (within threshold)
    #         initialize empty lists distance_groups and group_indices
    #         for each distance in nearest_distances:
    #             if no existing group within threshold:
    #                 group_indices_for_distance = find_indices_within_threshold(nearest_distances, distance)
    #                 add group_indices_for_distance to group_indices
    #                 add nearest_distances[group_indices_for_distance] to distance_groups

    #         # Step 5: Precompute the middle vector and internal average cosine angle for each group
    #         initialize empty dictionaries group_middle_vectors and group_internal_cos_angles
    #         for each group in group_indices:
    #             if group has multiple points:
    #                 vectors = compute_vectors_to_atom(i, group)
    #                 unit_vectors = normalize_each_vector(vectors)
    #                 middle_vector = sum(unit_vectors)
                    
    #                 if middle_vector is approximately zero:
    #                     set middle_vector to [0.0, 0.0, 0.0]
                    
    #                 # Calculate the internal squared average angle and take cosine
    #                 internal_cos_angle = compute_internal_cos_angle(vectors)
                    
    #                 store middle_vector in group_middle_vectors
    #                 store internal_cos_angle in group_internal_cos_angles
    #             else:
    #                 # For single atom group
    #                 store vector in group_middle_vectors
    #                 set internal_cos_angle to 1 in group_internal_cos_angles

    #         # Step 6: Calculate cosine angles between neighbors
    #         initialize empty list cos_angles
    #         for each pair (j, k) of neighbors up to n_nbr:
    #             if j and k are in the same group:
    #                 cos_angle = group_internal_cos_angles for that group
    #             else if j or k belongs to a group with a middle vector:
    #                 if middle vector is zero:
    #                     cos_angle = -100  # Padding value for undefined angles
    #                 else:
    #                     cos_angle = cosine of angle between middle vector and vector to other neighbor
    #             else:
    #                 cos_angle = cosine of angle between vectors to both neighbors

    #             if cos_angle is undefined:
    #                 cos_angle = -100  # Apply padding if undefined

    #             append cos_angle to cos_angles

    #         append cos_angles to cos_angles_list

    #     return cos_angles_list

    @staticmethod
    def get_nearest_neighbors_cos_angles(structure, n_nbr=8, distance_threshold=1e-3):

        def calculate_cosine_angle(v1, v2):
            cosine_angle = np.dot(v1, v2) / (np.linalg.norm(v1) * np.linalg.norm(v2))
            cosine_angle = np.clip(cosine_angle, -1, 1)  # Adjust for numerical stability
            return cosine_angle

        distance_matrix = structure.distance_matrix
        cos_angles_list = []

        for i in range(len(structure.sites)):
            # Step 1: Sort all neighbors by distance
            all_indices_sorted = np.argsort(distance_matrix[i])
            all_distances_sorted = distance_matrix[i, all_indices_sorted]

            # Exclude the current point itself
            all_indices_sorted = all_indices_sorted[all_indices_sorted != i]
            all_distances_sorted = all_distances_sorted[1:]  # Skip the first element (self)

            # Step 2: Determine the distance of the nth nearest neighbor
            n_nbr_distance = all_distances_sorted[n_nbr - 1]  # Distance of the nth neighbor

            # Step 3: Extend neighbors if further points have a distance within the threshold
            extended_count = n_nbr
            while extended_count < len(all_distances_sorted) and abs(all_distances_sorted[extended_count] - n_nbr_distance) < distance_threshold:
                extended_count += 1

            # Step 4: Select the extended list of nearest neighbors
            nearest_neighbors_indices = all_indices_sorted[:extended_count]
            nearest_distances = all_distances_sorted[:extended_count]

            # Step 5: Group neighbors by identical distances (within threshold)
            distance_groups = []
            group_indices = []
            
            for j in range(len(nearest_distances)):
                if not any(abs(nearest_distances[j] - group[0]) < distance_threshold for group in distance_groups):
                    indices = np.where(abs(nearest_distances - nearest_distances[j]) < distance_threshold)[0]
                    distance_groups.append(nearest_distances[indices])
                    group_indices.append(indices)
            
            # Step 6: Precompute the cumulative middle vector and internal average cosine for each group
            group_middle_vectors = {}
            group_internal_cos_angles = {}

            for idx, group in enumerate(group_indices):
                if len(group) > 1:
                    vectors = np.array([structure[nearest_neighbors_indices[g]].coords - structure[i].coords for g in group])
                    
                    # Normalize and accumulate vectors to get an average direction
                    unit_vectors = np.array([vec / np.linalg.norm(vec) for vec in vectors])
                    middle_vector = np.sum(unit_vectors, axis=0)
                    
                    # Calculate internal squared average angle
                    angles = []
                    for m in range(len(vectors)):
                        for n in range(m + 1, len(vectors)):
                            angle = np.arccos(calculate_cosine_angle(vectors[m], vectors[n]))
                            angles.append(angle ** 2)
                    avg_squared_angle = np.sqrt(np.mean(angles))
                    group_internal_cos_angles[idx] = np.cos(avg_squared_angle)  # Cosine of internal average angle
                    group_middle_vectors[idx] = middle_vector if np.linalg.norm(middle_vector) > 1e-6 else np.array([0.0, 0.0, 0.0])
                else:
                    group_middle_vectors[idx] = structure[nearest_neighbors_indices[group[0]]].coords - structure[i].coords
                    group_internal_cos_angles[idx] = 1  # Single atom, angle is zero

            # Step 7: Calculate cosines using the modified logic with cumulative middle vectors and internal angles
            cos_angles = []
            total_expected_angles = sum(range(n_nbr))  # Total expected angles for n_nbr neighbors

            for j in range(n_nbr - 1):  # Ensure n_nbr - 1 loops for dynamic padding
                actual_angles_this_round = 0
                for k in range(j + 1, n_nbr):
                    if j < len(nearest_neighbors_indices) and k < len(nearest_neighbors_indices):
                        group_j = next((idx for idx, g in enumerate(group_indices) if j in g), None)
                        group_k = next((idx for idx, g in enumerate(group_indices) if k in g), None)
                        
                        if group_j is not None and group_k is not None:
                            if group_j == group_k:
                                cos_angle = group_internal_cos_angles[group_j]
                            else:
                                cos_angle = -100 if np.linalg.norm(group_middle_vectors[group_j]) < 1e-6 or np.linalg.norm(group_middle_vectors[group_k]) < 1e-6 else calculate_cosine_angle(group_middle_vectors[group_j], group_middle_vectors[group_k])
                        elif group_j is not None:
                            vec_k = structure[nearest_neighbors_indices[k]].coords - structure[i].coords
                            cos_angle = -100 if np.linalg.norm(group_middle_vectors[group_j]) < 1e-6 else calculate_cosine_angle(group_middle_vectors[group_j], vec_k)
                        elif group_k is not None:
                            vec_j = structure[nearest_neighbors_indices[j]].coords - structure[i].coords
                            cos_angle = -100 if np.linalg.norm(group_middle_vectors[group_k]) < 1e-6 else calculate_cosine_angle(vec_j, group_middle_vectors[group_k])
                        else:
                            vec1 = structure[nearest_neighbors_indices[j]].coords - structure[i].coords
                            vec2 = structure[nearest_neighbors_indices[k]].coords - structure[i].coords
                            cos_angle = calculate_cosine_angle(vec1, vec2)
                        
                        cos_angles.append(cos_angle)
                        actual_angles_this_round += 1
                    else:
                        cos_angles.append(-100)  # Padding for missing neighbors

                expected_angles_this_round = n_nbr - 1 - j
                missing_angles_this_round = expected_angles_this_round - actual_angles_this_round
                cos_angles.extend([-100] * missing_angles_this_round)

            total_actual_angles = len(cos_angles)
            if total_actual_angles < total_expected_angles:
                cos_angles.extend([-100] * (total_expected_angles - total_actual_angles))
            
            cos_angles_list.append(cos_angles)

        return cos_angles_list

    def get_graph(self, cif_id):
        if self.read_from_pickle:
            file_structure = os.path.join(self.pickle_path, f"{cif_id}.pickle")
            with open(file_structure, "rb") as f:
                crystal_data = pickle.load(f)
        else:#generate in real time
            file_cif = os.path.join(self.cif_path, f"{cif_id}.cif")
            structure=self.read_structure(file_cif)
            structure=self.symmetry_operation(structure)
            if structure.num_sites > 1023:
                print("cif",cif_id,": Atom length {structure.num_sites} > 1023, please remove it, will not build its graph")
                return None
            else:
                atom_num = [site.specie.Z for site in structure]
                angles_list= self.get_nearest_neighbors_cos_angles(structure=structure, n_nbr=8)
                crystal_data={
                    "structure":structure,
                    "atom_num":atom_num,
                    "angles_list":angles_list,
                    "cif_id":cif_id,
                }


        graphdata = self.shuffle_structure_and_get_graph(crystal_data)
        # expand nbr_fea
        if self.if_conv:
            graphdata["nbr_fea"]=self.get_gaussian_distance(graphdata["nbr_fea"], num_step=self.nbr_fea_len).float()
        # # get image
        # if self.if_image:
        #     image_path = os.path.join(self.image_path, f"{cif_id}.npy")
        #     image_data  = np.load(image_path, allow_pickle=True)
        #     # Extract the image array and numerical values
        #     image = image_data[0]
        #     max_birth_1d = float(image_data[1])
        #     max_persistence_1d = float(image_data[2])
        #     max_birth_2d = float(image_data[3])
        #     max_persistence_2d = float(image_data[4])
        #     if image.shape[0] == 3:
        #         graphdata["image"] = torch.from_numpy(image[1:, :, :]).float()
        #     else:
        #         graphdata["image"] = torch.from_numpy(image).float()
        #     # Creating tensors from scalar values
        #     graphdata["max_birth_1d"] = max_birth_1d
        #     graphdata["max_persistence_1d"] = max_persistence_1d
        #     graphdata["max_birth_2d"] = max_birth_2d
        #     graphdata["max_persistence_2d"] = max_persistence_2d
        return graphdata  
    
    def get_image(self, cif_id):
        image_path = os.path.join(self.image_path, f"{cif_id}.npy")
        image_data  = np.load(image_path, allow_pickle=True)
        # Extract the image array and numerical values
        image = image_data[0]
        max_birth_1d = float(image_data[1])
        max_persistence_1d = float(image_data[2])
        max_birth_2d = float(image_data[3])
        max_persistence_2d = float(image_data[4])

        imagedata = dict()
        if image.shape[0] == 3:
            imagedata["image"] = torch.from_numpy(image[1:, :, :]).float()
        else:
            imagedata["image"] = torch.from_numpy(image).float()
        # Creating tensors from scalar values
        imagedata["max_birth_1d"] = max_birth_1d
        imagedata["max_persistence_1d"] = max_persistence_1d
        imagedata["max_birth_2d"] = max_birth_2d
        imagedata["max_persistence_2d"] = max_persistence_2d
        return imagedata      

    @staticmethod
    def make_grid_data(grid_data, emin=-5000.0, emax=5000, bins=101):
        """
        make grid_data within range (emin, emax) and
        make bins with logit function
        and digitize (0, bins)
        ****
            caution : 'zero' should be padding !!
            when you change bins, heads.MPP_heads should be changed
        ****
        """
        grid_data[grid_data <= emin] = emin
        grid_data[grid_data > emax] = emax

        x = np.linspace(emin, emax, bins)
        new_grid_data = np.digitize(grid_data, x) + 1

        return new_grid_data

    @staticmethod
    def calculate_volume(a, b, c, angle_a, angle_b, angle_c):
        a_ = np.cos(angle_a * np.pi / 180)
        b_ = np.cos(angle_b * np.pi / 180)
        c_ = np.cos(angle_c * np.pi / 180)

        v = a * b * c * np.sqrt(1 - a_**2 - b_**2 - c_**2 + 2 * a_ * b_ * c_)

        return v.item() / (60 * 60 * 60)  # normalized volume

    def get_raw_grid_data(self, cif_id):
        file_grid = os.path.join(self.grid_path, f"{cif_id}.grid")
        file_griddata = os.path.join(self.grid_path, f"{cif_id}.griddata16")

        # get grid
        with open(file_grid, "r") as f:
            lines = f.readlines()
            a, b, c = [float(i) for i in lines[0].split()[1:]]
            angle_a, angle_b, angle_c = [float(i) for i in lines[1].split()[1:]]
            cell = [int(i) for i in lines[2].split()[1:]]

        volume = self.calculate_volume(a, b, c, angle_a, angle_b, angle_c)

        # get grid data
        grid_data = pickle.load(open(file_griddata, "rb"))
        grid_data = self.make_grid_data(grid_data)
        grid_data = torch.FloatTensor(grid_data)

        return cell, volume, grid_data

    def get_grid_data(self, cif_id, draw_false_grid=False):
        cell, volume, grid_data = self.get_raw_grid_data(cif_id)
        ret = {
            "cell": cell,
            "volume": volume,
            "grid_data": grid_data,
        }

        if draw_false_grid:
            random_index = random.randint(0, len(self.cif_ids) - 1)
            cif_id = self.cif_ids[random_index]
            cell, volume, grid_data = self.get_raw_grid_data(cif_id)
            ret.update(
                {
                    "false_cell": cell,
                    "fale_volume": volume,
                    "false_grid_data": grid_data,
                }
            )
        return ret   


    def __getitem__(self, index):
        crystal_data = dict()
        cif_id = self.cif_ids[index]

        temp_graph_data = self.get_graph(cif_id)
        crystal_data.update(temp_graph_data if temp_graph_data is not None else print(f"{cif_id} has more than 1023 atoms, please use cif_to_data.py to rule out structures with more than 1023 atoms")) #get graph data and other labels

        if self.if_image:
            crystal_data.update(self.get_image(cif_id)) # get image data
        if self.if_grid:
            crystal_data.update(self.get_grid_data(cif_id)) # get grid data


        if self.downstream:
            downstream_data=self.id_downstream_data[cif_id]
            crystal_data.update({"target": downstream_data,})
        
        if "cdp" in self.tasks:
            density = float(self.id_density_data[cif_id])
            crystal_data.update({"density": density,})
            
        if "sgp" in self.tasks:    
            space_group = int(self.id_space_group_data[cif_id])
            crystal_data.update({"space_group": space_group,})
        
        if "sep" in self.tasks:    
            space_group = self.id_space_group_data[cif_id]
            symm_elem = self.symm_elem_data[space_group]
            crystal_data.update({"symm_elem": symm_elem,})
        """
        crystal_data = {
            "atom_num":atom_num,                           (torch.Tensor),     [n,1]
            "nbr_fea":nbr_fea,                             (torch.Tensor),     [n,num_nbr,self.nbr_fea_len]
            "nbr_fea_idx":nbr_fea_idx,                     (torch.LongTensor), [n,12]
            "distance_matrix":distance_matrix,             (torch.Tensor),     [n,n]
            "angles_list":angles_list,                     (torch.Tensor),     [n,n_nbr*(n_nbr-1)/2]
            "cif_id":cif_id,                               (str),  
            "density"                                      (float),
            "space_group"                                  (int),
            "symm_elem"                                    (list),             [22]
            / {downstream}                               }
        """

        return crystal_data

    @staticmethod
    def collate(batch,max_graph_len: int=1023,angle_nbr:int=8,mask_probability:float=0.15,tasks=[],if_conv=False,pos_emb="relative",if_alignn=True):
        """
        collate batch
        """
        batch_size = len(batch)
        keys = set([key for b in batch for key in b.keys()])

        dict_batch = {k: [dic[k] if k in dic else None for dic in batch] for k in keys}

        # image
        if 'image' in dict_batch:
            dict_batch["image"] = torch.stack(dict_batch["image"], dim=0)
            # dict_batch["max_birth_1d"] = torch.stack(dict_batch["max_birth_1d"], dim=0)
            # dict_batch["max_persistence_1d"] = torch.stack(dict_batch["max_persistence_1d"], dim=0)
            # dict_batch["max_birth_2d"] = torch.stack(dict_batch["max_birth_2d"], dim=0)
            # dict_batch["max_persistence_2d"] = torch.stack(dict_batch["max_persistence_2d"], dim=0)
        if "grid_data" in dict_batch:
            from torch.nn.functional import interpolate
            batch_grid_data = dict_batch["grid_data"]
            batch_cell = dict_batch["cell"]
            new_grids = []

            for bi in range(batch_size):
                orig = batch_grid_data[bi].view(batch_cell[bi][::-1]).transpose(0, 2)
                if batch_cell[bi] == [30, 30, 30]:  # version >= 1.1.2
                    orig = orig[None, None, :, :, :]
                else:
                    orig = interpolate(
                        orig[None, None, :, :, :],
                        size=[30, 30, 30],
                        mode="trilinear",
                        align_corners=True,
                    )
                new_grids.append(orig)
            new_grids = torch.concat(new_grids, axis=0)
            dict_batch["grid"] = new_grids

            if "false_grid_data" in dict_batch.keys():
                batch_false_grid_data = dict_batch["false_grid_data"]
                batch_false_cell = dict_batch["false_cell"]
                new_false_grids = []
                for bi in range(batch_size):
                    orig = batch_false_grid_data[bi].view(batch_false_cell[bi])
                    if batch_cell[bi] == [30, 30, 30]:  # version >= 1.1.2
                        orig = orig[None, None, :, :, :]
                    else:
                        orig = interpolate(
                            orig[None, None, :, :, :],
                            size=[30, 30, 30],
                            mode="trilinear",
                            align_corners=True,
                        )
                    new_false_grids.append(orig)
                new_false_grids = torch.concat(new_false_grids, axis=0)
                dict_batch["false_grid"] = new_false_grids

            dict_batch.pop("grid_data", None)
            dict_batch.pop("false_grid_data", None)
            dict_batch.pop("cell", None)
            dict_batch.pop("false_cell", None)            

        #atom graph for cgcnn
        batch_atom_num = dict_batch["atom_num"]

        #for cgcnn
        if if_conv:
            batch_nbr_idx = dict_batch["nbr_fea_idx"]
            batch_nbr_fea = dict_batch["nbr_fea"]
            crystal_atom_idx = []
            base_idx = 0
            for i, nbr_idx in enumerate(batch_nbr_idx):
                n_i = nbr_idx.shape[0]
                crystal_atom_idx.append(torch.arange(n_i) + base_idx)
                nbr_idx += base_idx
                base_idx += n_i
            dict_batch["nbr_fea_idx"] = torch.cat(batch_nbr_idx, dim=0)
            dict_batch["nbr_fea"] = torch.cat(batch_nbr_fea, dim=0)
        else:
            crystal_atom_idx = []
            base_idx = 0
            for i, nbr_idx in enumerate(batch_atom_num):
                n_i = nbr_idx.shape[0]
                crystal_atom_idx.append(torch.arange(n_i) + base_idx)
                base_idx += n_i

        dict_batch["atom_num"] = torch.cat(batch_atom_num, dim=0)
        dict_batch["crystal_atom_idx"] = crystal_atom_idx

        #dgl graph 
        if if_alignn:
            dict_batch["graph"] = dgl.batch(dict_batch["graph"])
            dict_batch["line_graph"] = dgl.batch(dict_batch["line_graph"])
            
        #for apc task
        if "apc" in tasks:
            batch_atom_pairs = dict_batch["atom_pairs"]
            batch_ap_labels = dict_batch["ap_labels"]
            dict_batch["atom_pairs"] = torch.stack(batch_atom_pairs)  # [B, m, 2]
            dict_batch["ap_labels"] = torch.stack(batch_ap_labels)    # [B, m]

        #for adp task
        if "adp" in tasks:
            batch_dist_pairs = dict_batch["dist_pairs"]
            batch_dist_labels = dict_batch["dist_labels"]
            dict_batch["dist_pairs"] = torch.stack(batch_dist_pairs)  # [B, n_dist, 2]
            dict_batch["dist_labels"] = torch.stack(batch_dist_labels)    # [B, n_dist]

        #for aap task
        if "aap" in tasks:
            batch_angle_pairs = dict_batch["angle_pairs"]
            batch_angle_labels = dict_batch["angle_labels"]
            dict_batch["angle_pairs"] = torch.stack(batch_angle_pairs)  # [B, n_angle, 2]
            dict_batch["angle_labels"] = torch.stack(batch_angle_labels)    # [B, n_angle]

        #for map task
        # mask atoms
        def apply_mask_to_atom_num(atom_num, mask_probability=0.15):
            """
            Applies masking logic directly to atom numbers in the dataset.
            Returns:
            masked_atom_num (torch.Tensor): Tensor of atom numbers after applying mask.
            atom_label (torch.Tensor): Labels tensor for loss computation, with -1 indicating atoms to ignore.
            """
            atom_label = torch.full_like(atom_num, -1)
            mask = torch.rand(atom_num.size(0)) < mask_probability
            masked_atom_num = atom_num.clone()

            for i in range(masked_atom_num.size(0)):
                if mask[i]:
                    atom_label[i] = atom_num[i]# Set the correct label for masked atom
                    decision = torch.rand(1).item()
                    if decision < 0.8:  # 80% chance to replace with [MASK], in this case 0
                        masked_atom_num[i] = 0
                    elif decision < 0.9:  # 10% chance for random replacement
                        masked_atom_num[i] = torch.randint(1, 119, (1,)).item()
                    # Remaining 10% chance to keep unchanged
                    
            return masked_atom_num, atom_label 
        if "map" in tasks:
            masked_atom_num, atom_label = apply_mask_to_atom_num(dict_batch["atom_num"],mask_probability)
            dict_batch["atom_num"]=masked_atom_num
            padded_atom_label = torch.full((batch_size, max_graph_len), -1)
            for i, idx in enumerate(crystal_atom_idx):
                    length = len(idx)
                    padded_atom_label[i, :length] = atom_label[idx]
            dict_batch["atom_label"]=padded_atom_label
        
        # for positional embedding
        if pos_emb == "relative":
            #distance
            batch_distance_matrix = dict_batch["distance_matrix"]
            padded_distance_matrices = torch.zeros(batch_size, max_graph_len, max_graph_len)
            mask_distance_matrices = torch.zeros(batch_size, max_graph_len, max_graph_len)
            for i, matrix in enumerate(batch_distance_matrix):
                n = matrix.size(0)
                padded_distance_matrices[i, :n, :n] = matrix
                mask_distance_matrices[i, :n, :n] = 1
            dict_batch["padded_distance_matrices"] = padded_distance_matrices
            dict_batch["mask_distance_matrices"] = mask_distance_matrices

            #angle
            def generate_indices(angle_nbr):
                indices = []
                start = 0
                for i in range(7, 0, -1):
                    end = start + i
                    indices.extend(range(start, start + angle_nbr-1))
                    start = end
                    angle_nbr -= 1
                    if angle_nbr <= 0:
                        break
                return indices
            #select angles based on angle_nbr
            batch_angles_list= dict_batch["angles_list"] 
            angle_indices=generate_indices(angle_nbr)
            for i, tensor in enumerate(batch_angles_list):
                batch_angles_list[i]=tensor[:, angle_indices]

            x_dim = batch_angles_list[0].shape[1]
            padded_angle_matrices = torch.full((batch_size, max_graph_len, x_dim), -100.0)
            mask_angle_matrices = torch.full((batch_size, max_graph_len, x_dim), 1)
            for i, matrix in enumerate(batch_angles_list):
                n = matrix.shape[0]
                padded_angle_matrices[i, :n, :] = matrix

            mask_angle_matrices[padded_angle_matrices == -100.0] = 0
            dict_batch["padded_angle_matrices"] = padded_angle_matrices
            dict_batch["mask_angle_matrices"] = mask_angle_matrices   

        elif pos_emb == "absolute":
            batch_abs_pos = dict_batch["abs_pos"]
            pos_dim = batch_abs_pos[0].shape[1]
            padded_abs_pos = torch.zeros(batch_size, max_graph_len, pos_dim)
            for i, matrix in enumerate(batch_abs_pos):
                n = matrix.shape[0]
                padded_abs_pos[i, :n, :] = matrix
            dict_batch["padded_abs_pos"] = padded_abs_pos

        elif pos_emb=='both':
            #distance
            batch_distance_matrix = dict_batch["distance_matrix"]
            padded_distance_matrices = torch.zeros(batch_size, max_graph_len, max_graph_len)
            mask_distance_matrices = torch.zeros(batch_size, max_graph_len, max_graph_len)
            for i, matrix in enumerate(batch_distance_matrix):
                n = matrix.size(0)
                padded_distance_matrices[i, :n, :n] = matrix
                mask_distance_matrices[i, :n, :n] = 1
            dict_batch["padded_distance_matrices"] = padded_distance_matrices
            dict_batch["mask_distance_matrices"] = mask_distance_matrices

            #angle
            def generate_indices(angle_nbr):
                indices = []
                start = 0
                for i in range(7, 0, -1):
                    end = start + i
                    indices.extend(range(start, start + angle_nbr-1))
                    start = end
                    angle_nbr -= 1
                    if angle_nbr <= 0:
                        break
                return indices
            #select angles based on angle_nbr
            batch_angles_list= dict_batch["angles_list"] 
            angle_indices=generate_indices(angle_nbr)
            for i, tensor in enumerate(batch_angles_list):
                batch_angles_list[i]=tensor[:, angle_indices]

            x_dim = batch_angles_list[0].shape[1]
            padded_angle_matrices = torch.full((batch_size, max_graph_len, x_dim), -100.0)
            mask_angle_matrices = torch.full((batch_size, max_graph_len, x_dim), 1)
            for i, matrix in enumerate(batch_angles_list):
                n = matrix.shape[0]
                padded_angle_matrices[i, :n, :] = matrix

            mask_angle_matrices[padded_angle_matrices == -100.0] = 0
            dict_batch["padded_angle_matrices"] = padded_angle_matrices
            dict_batch["mask_angle_matrices"] = mask_angle_matrices  
            #absolute
            batch_abs_pos = dict_batch["abs_pos"]
            pos_dim = batch_abs_pos[0].shape[1]
            padded_abs_pos = torch.zeros(batch_size, max_graph_len, pos_dim)
            for i, matrix in enumerate(batch_abs_pos):
                n = matrix.shape[0]
                padded_abs_pos[i, :n, :] = matrix
            dict_batch["padded_abs_pos"] = padded_abs_pos
        else:
            pass

        return dict_batch



import os
from sacred import Experiment
from MCRT.modules.utils import _set_load_path, _loss_names
from MCRT import __root_dir__

ex = Experiment("pretrained_MCRT", save_git_info=False)

@ex.config
def config():
    # model
    exp_name = "pretraining_MCRT"
    seed = 0
    loss_names = _loss_names({
        "map": 0,
        "apc": 0,  
        "sgp": 0,
        "sep": 0,
        "cdp": 0,
        "adp": 0,
        "aap": 0,
        "ucp": 0,})
    visualize=True
    # train_val_test_split=[0.8, 0.1, 0.1]
    # dataset_seed=123

    max_graph_len = 1023  # number of maximum atoms in P1 unit cell
    read_from_pickle = True # if read structure and angle list from pickle, if False, read cif and generate it in real time

    # graph setting, for cgcnn
    if_conv = False #for cgcnn, the data load method for cgcnn is different from other GNNs.
    max_num_nbr = 12
    atom_fea_len = 64
    nbr_fea_len = 41
    n_conv = 3
    mask_probability = 0.15 # proportion of masked atoms

    # for alignn
    if_alignn = True
    alignn_num_conv = 4
    alignn_hidden_dim = 256
    alignn_rbf_distance_dim = 40 # RDF expansion dimension for edge distance
    alignn_rbf_triplet_dim = 40 # RDF expansion dimension for triplet angle
    alignn_batch_norm = True
    alignn_dropout = 0.0
    alignn_residual = False

    # chemical positional embedding 
    angle_nbr = 8 # number of nbr atoms when generating the angle matrix, int between 2-8
    pos_emb = "relative" #  "absolute" or "relative" or "both"
    # pretext tasks
    num_ap = 200 # number of atom pairs from each crystal, can't be odd
    n_dist = 100 # number of atom pairs for distance prediction
    n_angle = 100 # number of atom pairs for angle prediction

    # sep_weights = [4.43535968,68.57053843,76.72758503,99.97791198,50.52052353,9.96930181,2.19511,78.12710143,89.54104273,90.02648945,91.84479127,66.14665125,87.25459766,81.96008386,100.35775213,94.87247941,99.90788809,92.78795367,100.04803458,95.63538094,35.27829775,2.21395585] # generated by weights_log = 1 / np.log(epsilon=1.01 + frequencies)
    #sep_weights = [1,1,1,1,1,1,1]
    sep_weights = [9.58304127,3.39176836,8.91170359,5.50722613,1.93331989,9.76274322,1.97087159]# generated by weights_log = 1 / np.log(epsilon=1.1 + frequencies)
    # transformer setting
    num_blocks = 12
    hid_dim = 768
    num_heads = 12
    mlp_ratio = 4
    qkv_bias = True
    qk_scale = None
    drop_rate = 0.1
    attn_drop = 0.0
    drop_path_rate = 0.1

    # for persistence images
    if_image = True #if use image modality
    img_size=(50,50)
    patch_size=(5,5)
    in_chans=1

    # for energy grid
    if_grid = False # if use energy grid, for moftransformer only
    grid_img_size = 30
    grid_patch_size = 5  # length of patch
    grid_in_chans = 1  # channels of grid image

    # downstream
    downstream = ""
    n_classes = 0

    # Optimizer Setting
    optim_type = "adamw"  # adamw, adam, sgd (momentum=0.9)
    learning_rate = 1e-4
    weight_decay = 1e-2
    decay_power = (
        1  # default polynomial decay, [cosine, constant, constant_with_warmup]
    )
    max_epochs = 100
    max_steps = -1  # num_data * max_epoch // batch_size (accumulate_grad_batches)
    warmup_steps = 0.05  # int or float ( max_steps * warmup_steps)
    end_lr = 0
    lr_mult = 1  # multiply lr for downstream heads

    # PL Trainer Setting
    resume_from = None
    val_check_interval = 1.0
    test_only = False
    test_to_csv = True
    cls_to_csv = False

    # below params varies with the environment
    root_dataset = os.path.join(__root_dir__, "cifs","test")
    log_dir = "logs/"
    batch_size = 512  # desired batch size; for gradient accumulation
    per_gpu_batchsize = 32  # you should define this manually with per_gpu_batch_size
    accelerator = "auto"
    devices = "auto"
    num_nodes = 1
    strategy = "ddp"
    load_path = _set_load_path(None) # path must be 'MCRT' (for finetune), None (for pretrain), or path to *.ckpt (for using finetuned model)

    num_workers = 6  # the number of cpu's core
    precision = '32-true' 
    """
    precision (Union[Literal[64, 32, 16], Literal['transformer-engine', 'transformer-engine-float16', '16-true', '16-mixed', 'bf16-true', 'bf16-mixed', '32-true', '64-true'], 
    Literal['64', '32', '16', 'bf16'], None]) , Double precision (64, '64' or '64-true'), 
    full precision (32, '32' or '32-true'), 16bit mixed precision (16, '16', '16-mixed') or bfloat16 mixed precision ('bf16', 'bf16-mixed'). 
    """

    # normalization target
    mean = None
    std = None


@ex.named_config
def example():
    exp_name = "example"
    root_dataset = "examples/dataset"
    downstream = "example"
    max_epochs = 20
    batch_size = 32


@ex.named_config
def test():
    exp_name = "test"
    loss_names = _loss_names({"regression": 0})
    root_dataset = "examples/dataset"
    downstream = ""
    max_epochs = 20
    batch_size = 32


@ex.named_config
def total_h2_uptake():
    exp_name = "total_h2_uptake"
    # root_dataset = ##
    downstream = "h2_uptake"
    max_epochs = 20
    batch_size = 32
    mean = 423.251
    std = 136.293


@ex.named_config
def total_h2_wc():
    exp_name = "total_h2_wc"
    # root_dataset = ##
    downstream = "h2_wc"
    max_epochs = 20
    batch_size = 32
    mean = 241.300
    std = 152.052


@ex.named_config
def qmof_bandgap():
    exp_name = "qmof_bandgap"
    # root_dataset = ##
    downstream = "qmof_bandgap"
    max_epochs = 20
    batch_size = 32
    mean = 2.086
    std = 1.131



"""
project distance matrix and angle matrix to hid_dim
"""
import torch
import torch.nn as nn
import math
from MCRT.data_processor.dataset import Dataset

def main():
    dataset_test=Dataset(r'D:\Projects\MyProjects\MCRT\MCRT\cifs\test','train',1023,4)
    batch=[dataset_test[i] for i in range(32)]
    collated_batch=dataset_test.collate(batch)
    print(collated_batch["padded_angle_matrices"])
    angle_MLP_Projector=MLP_Projector(
        in_features=6,
        hidden_features=512,
        out_features=512,
        act_layer=nn.GELU,
        drop=0.0,)
    angleoutput=angle_MLP_Projector(collated_batch["padded_angle_matrices"],collated_batch["mask_angle_matrices"])
    print(angleoutput.shape)

    dist_MLP_Projector=MLP_Projector(
        in_features=1023,
        hidden_features=512,
        out_features=512,
        act_layer=nn.GELU,
        drop=0.0,)
    output=dist_MLP_Projector(collated_batch["padded_distance_matrices"],collated_batch["mask_distance_matrices"])
    print(output.shape)
    

class MLP_Projector(nn.Module):
    def __init__(
        self,
        in_features,
        hidden_features=None,
        out_features=None,
        act_layer=nn.GELU,
        drop=0.0,
    ):
        super().__init__()
        out_features = out_features or in_features
        hidden_features = hidden_features or in_features
        self.fc1 = nn.Linear(in_features, hidden_features)
        self.act = act_layer()
        self.fc2 = nn.Linear(hidden_features, out_features)
        self.drop = nn.Dropout(drop)

    def forward(self, x, mask):
        x = self.fc1(x)
        x = self.act(x)
        x = self.drop(x)
        x = self.fc2(x)
        x = self.drop(x)
        if mask.dim() == 3:
            mask = mask.any(dim=-1, keepdim=True)
        x = x * mask
        return x



class PositionalEncoding3D(nn.Module):
    def __init__(self, embed_dim, dropout=0.1):
        super(PositionalEncoding3D, self).__init__()
        self.dropout = nn.Dropout(p=dropout)
        self.embed_dim = embed_dim

    def forward(self, x, pos):
        pos = pos * 10
        device = x.device
        div = torch.exp(torch.arange(0., self.embed_dim, 2) * -(math.log(10000.0) / self.embed_dim)).to(device).type(x.dtype)
        for i in range(3):
            pe = torch.zeros(x.shape, device=device, dtype=x.dtype)
            pe[..., 0::2] = torch.sin(pos[..., i].unsqueeze(-1) * div)
            pe[..., 1::2] = torch.cos(pos[..., i].unsqueeze(-1) * div)
            x += pe
        return self.dropout(x)

class PositionalEncoding3D_new(nn.Module):
    def __init__(self, embed_dim, dropout=0.1):
        super(PositionalEncoding3D_new, self).__init__()
        self.dropout = nn.Dropout(p=dropout)
        self.embed_dim = embed_dim
        self.pe_linear=nn.Linear(3*embed_dim,embed_dim)

    def forward(self, x, pos):
      
        pos = pos * 10

       
        div = torch.exp(torch.arange(0., self.embed_dim, 2) * -(math.log(10000.0) / self.embed_dim)).double().cuda()
        for i in range(3):
            pe = torch.zeros(x.shape).cuda()
            pe[..., 0::2] = torch.sin(pos[..., i].unsqueeze(2) * div)
            pe[..., 1::2] = torch.cos(pos[..., i].unsqueeze(2) * div)
            if i==0:
                pe_all=pe
            else:
                pe_all=torch.cat((pe_all,pe),dim=-1)
        pe_final=self.pe_linear(pe_all)
        x += pe_final
        return self.dropout(x)

if __name__ == "__main__":
    main()  